# -*- coding: utf-8 -*-

"""This module provides classes related to infra, containers, and networking.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from .containercluster import *
from .network import *
from .secureconnectgateway import *



# -*- coding: utf-8 -*-

"""This module provides classes related to workload vulnerabilities.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from .vulnerability import *
from .vulnerabilityreport import *



import pytest

from illumio.infrastructure import ContainerWorkloadProfile, LabelRestriction
from illumio.util import EnforcementMode

from helpers import random_string


@pytest.fixture
def container_workload_profile(pce, session_identifier, container_cluster, env_label, loc_label):
    identifier = random_string()
    container_workload_profile = ContainerWorkloadProfile(
        name='{}-{}'.format(session_identifier, identifier),
        description='Created by illumio python library integration tests',
        managed=True,
        assign_labels=[env_label, loc_label],
        enforcement_mode=EnforcementMode.VISIBILITY_ONLY
    )
    container_workload_profile = pce.container_workload_profiles.create(
        container_workload_profile,
        parent=container_cluster.href
    )
    yield container_workload_profile
    pce.container_workload_profiles.delete(container_workload_profile)


def test_get_by_reference(pce, container_workload_profile):
    workload_profile = pce.container_workload_profiles.get_by_reference(container_workload_profile)
    assert workload_profile.href == container_workload_profile.href


def test_get_from_container_cluster(pce, container_cluster, container_workload_profile):
    container_workload_profiles = pce.container_workload_profiles.get(parent=container_cluster.href)
    assert len(container_workload_profiles) == 2  # every cluster has a Default profile


def test_update_container_workload_profile(pce, container_workload_profile, app_label, env_label, loc_label):
    label_restrictions = [
        LabelRestriction(key='app', restriction=[app_label]),
        LabelRestriction(key='env', assignment=env_label),
        LabelRestriction(key='loc', assignment=loc_label),
    ]
    pce.container_workload_profiles.update(container_workload_profile, {'labels': label_restrictions})
    updated_container_workload_profile = pce.container_workload_profiles.get_by_reference(container_workload_profile)
    assert len(updated_container_workload_profile.labels) == 3



# -*- coding: utf-8 -*-

"""This module provides classes related to IP list policy objects.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from ipaddress import ip_address, ip_network
from typing import List

from illumio import IllumioException
from illumio.util import JsonObject, MutableObject, pce_api


@dataclass
class IPRange(JsonObject):
    """Represents a range of one or more IP addresses in an IP list.

    Args:
        from_ip (str, optional): IP address at the start of the range. Can be a
            single IP or CIDR range, e.g. "10.0.0.0/8".
        to_ip (str, optional): IP address at the end of the range. If provided,
            ``from_ip`` must be a single IP address
        exclusion (bool, optional): if True, this range represents an exclusion
            rather than an inclusion in the IP list object.
        description (str, optional): optional description.

    Raises:
        IllumioException: if an invalid IP range is given.
    """
    from_ip: str = None
    to_ip: str = None
    exclusion: bool = None
    description: str = None

    def _validate(self):
        try:
            from_net = ip_network(self.from_ip)
            if self.to_ip:
                to_ip = ip_address(self.to_ip)
                if from_net.prefixlen < 32:
                    raise "Can't specify CIDR block and to_ip in same range"
                if to_ip <= from_net.network_address:
                    raise "to_ip address must be greater than from_ip address"
        except Exception as e:
            raise IllumioException("Invalid IP range: {}".format(e))
        return super()._validate()


@dataclass
class FQDN(JsonObject):
    """Represents a fully-qualified domain name associated with an IP list.

    Args:
        fqdn (str, optional): fully-qualified domain name.
        description (str, optional): optional description.
    """
    fqdn: str = None
    description: str = None


@dataclass
@pce_api('ip_lists', is_sec_policy=True)
class IPList(MutableObject):
    """Represents an IP list in the PCE.

    IP lists are list of IP addresses, subnets, CIDR blocks, and/or FQDNs.
    They can be used in conjunction with other security policy objects to
    allow or deny traffic from these defined ranges.

    See https://docs.illumio.com/core/21.5/Content/Guides/security-policy/security-policy-objects/ip-lists.htm

    Usage:
        >>> from illumio import PolicyComputeEngine, IPList, IPRange
        >>> pce = PolicyComputeEngine('my.pce.com')
        >>> pce.set_credentials('api_key_username', 'api_key_secret')
        >>> ip_list = IPList(
        ...     name='IPL-INTERNAL',
        ...     ip_ranges=IPRange(
        ...         from_ip='192.168.0.0/16'
        ...     )
        ... )
        >>> ip_list = pce.ip_lists.create(ip_list)
        >>> ip_list
        IPList(
            href='/orgs/1/sec_policy/draft/ip_lists/22',
            name='IPL-INTERNAL',
            ip_ranges=IPRange(
                from_ip='192.168.0.0/16'
            ),
            ...
        )
    """
    ip_ranges: List[IPRange] = None
    fqdns: List[FQDN] = None


__all__ = [
    'IPRange',
    'FQDN',
    'IPList',
]



import json
import os
import re
from typing import List

import pytest

from illumio.infrastructure import ContainerCluster

CONTAINER_CLUSTERS = os.path.join(pytest.DATA_DIR, 'container_clusters.json')


@pytest.fixture(scope='module')
def container_clusters() -> List[dict]:
    with open(CONTAINER_CLUSTERS, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(scope='module')
def container_cluster() -> ContainerCluster:
    return ContainerCluster(
        name='CC-TEST',
        description='Test cluster'
    )


@pytest.fixture(autouse=True)
def container_clusters_mock(pce_object_mock, container_clusters):
    pce_object_mock.add_mock_objects(container_clusters)


@pytest.fixture(autouse=True)
def mock_requests(requests_mock, get_callback, post_callback, put_callback, delete_callback):
    pattern = re.compile('/container_clusters')
    requests_mock.register_uri('GET', pattern, json=get_callback)
    requests_mock.register_uri('POST', pattern, json=post_callback)
    requests_mock.register_uri('PUT', pattern, json=put_callback)
    requests_mock.register_uri('DELETE', pattern, json=delete_callback)


def test_get_cluster_by_partial_name(pce):
    container_clusters = pce.container_clusters.get(params={'name': 'CC-'})
    assert len(container_clusters) == 2


def test_get_by_name(pce):
    container_cluster = pce.container_clusters.get_by_name('CC-EKS-PROD')
    assert container_cluster


def test_container_cluster_create(pce, container_cluster):
    container_cluster = pce.container_clusters.create(container_cluster)
    container_clusters = pce.container_clusters.get()
    assert len(container_clusters) == 3
    assert container_cluster.href in [p.href for p in container_clusters]


def test_container_cluster_update(pce):
    test_description = 'Updated test cluster'
    container_clusters = pce.container_clusters.get()
    container_cluster = container_clusters[0]
    container_cluster.description = test_description
    pce.container_clusters.update(container_cluster.href, container_cluster)
    container_cluster = pce.container_clusters.get_by_reference(container_cluster.href)
    assert container_cluster.description == test_description



import json
import os
import re
from typing import List

import pytest

from illumio.policyobjects import Service, ServicePort
from illumio.util import IllumioEncoder, convert_protocol, DRAFT, ACTIVE

SERVICES = os.path.join(pytest.DATA_DIR, 'services.json')
DEFAULT_SERVICE_NAME = 'All Services'


@pytest.fixture(scope='module')
def services() -> List[dict]:
    with open(SERVICES, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(scope='module')
def web_service() -> Service:
    return Service(
        name='S-Web',
        service_ports=[
            ServicePort(port=443, proto='tcp'),
            ServicePort(port=80, proto='tcp'),
        ]
    )


@pytest.fixture(autouse=True)
def services_mock(pce_object_mock, services):
    pce_object_mock.add_mock_objects(services)


@pytest.fixture(autouse=True)
def mock_requests(requests_mock, get_callback, post_callback, put_callback, delete_callback):
    pattern = re.compile('/sec_policy/(draft|active)/services')
    requests_mock.register_uri('GET', pattern, json=get_callback)
    requests_mock.register_uri('POST', pattern, json=post_callback)
    requests_mock.register_uri('PUT', pattern, json=put_callback)
    requests_mock.register_uri('DELETE', pattern, json=delete_callback)


def test_compare_service_ports(pce, web_service):
    http_service = pce.services.get(params={'name': 'S-HTTP', 'max_results': 1})[0]
    assert len(web_service.service_ports) == len(http_service.service_ports)
    assert set([(sp.port, sp.proto) for sp in web_service.service_ports]) == set([(sp.port, sp.proto) for sp in http_service.service_ports])


def test_proto_encoding():
    https_service_port = ServicePort(port=443, proto='tcp')
    assert https_service_port.proto == convert_protocol('tcp')
    encoded_https_service_port = json.loads(json.dumps(https_service_port, cls=IllumioEncoder))
    assert encoded_https_service_port == {'port': 443, 'proto': 6}


def test_get_default_service(pce):
    default_service = pce.services.get(params={'name': DEFAULT_SERVICE_NAME, 'max_results': 1}, policy_version=ACTIVE)[0]
    assert default_service.name == DEFAULT_SERVICE_NAME


def test_get_by_reference(pce):
    any_service = pce.services.get_by_reference('/orgs/1/sec_policy/active/services/1')
    assert any_service.name == DEFAULT_SERVICE_NAME


def test_get_by_partial_name(pce):
    services = pce.services.get(params={'name': 'S-'}, policy_version=DRAFT)
    assert len(services) == 2


def test_get_by_name(pce):
    service = pce.services.get_by_name('S-HTTP')
    assert service


def test_get_active_service(pce):
    services = pce.services.get(params={'name': 'S-'}, policy_version=ACTIVE)
    assert len(services) == 0


def test_create_service(pce, web_service):
    service = pce.services.create(web_service)
    assert service.href != ''
    fetched_service = pce.services.get_by_reference(service.href)
    assert fetched_service == service


def test_update_service(pce):
    service = pce.services.get(params={'name': 'S-HTTP', 'max_results': 1})[0]
    pce.services.update(service.href, {'service_ports': [ServicePort(port=8080, proto='tcp')]})
    updated_service = pce.services.get_by_reference(service.href)
    print(updated_service)
    assert len(updated_service.service_ports) > 0
    assert updated_service.service_ports[0].port == 8080



# -*- coding: utf-8 -*-

"""This module provides utility classes, functions, and constant variables.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from .constants import *
from .functions import *
from .jsonutils import *



# -*- coding: utf-8 -*-

"""This module provides common exception classes used throughout the application.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
class IllumioException(Exception):
    """Superclass for Illumio library exceptions."""


class IllumioApiException(IllumioException):
    """Superclass for exceptions generated by the Illumio API code."""


class IllumioIntegerValidationException(IllumioException):
    """Raised when invalid integer values are provided."""
    def __init__(self, val, minimum: int, maximum: int, message="") -> None:
        if not message:
            message = "Invalid value {} - must be integer between {} and {}".format(
                val, minimum, maximum
            )
        super().__init__(message)


__all__ = [
    'IllumioException',
    'IllumioApiException',
    'IllumioIntegerValidationException',
]



# -*- coding: utf-8 -*-

"""This module is a stub for vulnerability report objects.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass

from illumio.util import MutableObject


@dataclass
class VulnerabilityReport(MutableObject):
    pass


__all__ = [
    'VulnerabilityReport',
]



import pytest

from illumio.workloads import Workload

from helpers import random_string


@pytest.fixture
def workload(pce, session_identifier):
    identifier = random_string()
    hostname = '{}.{}'.format(session_identifier, identifier)
    workload = pce.workloads.create(
        Workload(
            name=hostname,
            hostname=hostname,
            description='Created by illumio python library integration tests',
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield workload
    pce.workloads.delete(workload.href)

__all__ = ['workload']



from illumio import PolicyComputeEngine


def test_pce_connection(pce: PolicyComputeEngine):
    assert pce.check_connection()


def test_invalid_org_id(pce: PolicyComputeEngine):
    # since pce is a session fixture, store the original id
    # to reset after we test the failed connection request
    configured_org_id = pce.org_id
    pce.org_id = 999
    assert not pce.check_connection()
    pce.org_id = configured_org_id



import pytest

from illumio.policyobjects import Service, ServicePort

from helpers import random_string


@pytest.fixture
def web_service(pce, session_identifier):
    identifier = random_string()
    service = pce.services.create(
        Service(
            name='{}-S-WEB-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests',
            service_ports=[
                ServicePort(port=80, proto='tcp'),
                ServicePort(port=443, proto='tcp')
            ],
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield service
    pce.services.delete(service.href)


@pytest.fixture
def well_known_service(pce, session_identifier):
    identifier = random_string()
    service = pce.services.create(
        Service(
            name='{}-S-WELL-KNOWN-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests',
            service_ports=[
                ServicePort(port=1, to_port=1023, proto='tcp')
            ],
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield service
    pce.services.delete(service.href)


@pytest.fixture
def rdp_service(pce, session_identifier):
    identifier = random_string()
    service = pce.services.create(
        Service(
            name='{}-S-RDP-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests',
            service_ports=[
                ServicePort(port=3389, proto='tcp'),
                ServicePort(port=3389, proto='udp')
            ],
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield service
    pce.services.delete(service.href)

__all__ = [
    'web_service',
    'well_known_service',
    'rdp_service'
]



import re
from collections import namedtuple
from dataclasses import fields

import pytest

from illumio import PolicyComputeEngine
from illumio.accessmanagement import User
from illumio.exceptions import IllumioIntegerValidationException
from illumio.infrastructure import ContainerWorkloadProfile
from illumio.policyobjects import Label
from illumio.rules import Rule
from illumio.util import PCE_APIS, DRAFT, ACTIVE

from mocks import MockResponse


def test_custom_protocol():
    pce = PolicyComputeEngine('http://my.pce.com')
    assert pce.base_url == 'http://my.pce.com:443/api/v2'


def test_invalid_protocol():
    pce = PolicyComputeEngine('ftp://my.pce.com')
    assert pce.base_url == 'https://my.pce.com:443/api/v2'


def test_protocol_parsing():
    pce = PolicyComputeEngine('httpslab.pce.com')
    assert pce.base_url == 'https://httpslab.pce.com:443/api/v2'


def test_path_parsing():
    pce = PolicyComputeEngine('my.pce.com/api/v2')
    assert pce.base_url == 'https://my.pce.com:443/api/v2'


def test_int_org_id():
    pce = PolicyComputeEngine('my.pce.com', org_id=1)
    assert pce.base_url == 'https://my.pce.com:443/api/v2'


def test_invalid_org_id():
    with pytest.raises(IllumioIntegerValidationException):
        PolicyComputeEngine('my.pce.com', org_id='invalid')


def test_org_id_lower_bound():
    with pytest.raises(IllumioIntegerValidationException):
        PolicyComputeEngine('my.pce.com', org_id=0)


def test_invalid_port():
    with pytest.raises(IllumioIntegerValidationException):
        PolicyComputeEngine('my.pce.com', port='invalid')


def test_port_lower_bound():
    with pytest.raises(IllumioIntegerValidationException):
        PolicyComputeEngine('my.pce.com', port=-1)


def test_port_upper_bound():
    with pytest.raises(IllumioIntegerValidationException):
        PolicyComputeEngine('my.pce.com', port=65536)


_API = namedtuple('API', [
    'name',
    'endpoint',
    'object_class',
    'is_sec_policy',
    'is_global'
])


@pytest.fixture(autouse=True)
def mock_requests(requests_mock):
    pattern = re.compile('/')
    requests_mock.register_uri('GET', pattern)
    requests_mock.register_uri('POST', pattern, status_code=201)
    requests_mock.register_uri('PUT', pattern, status_code=204)
    requests_mock.register_uri('DELETE', pattern, status_code=204)


@pytest.mark.parametrize(
    "api,policy_version,parent,expected", [
        (
            _API(name='users', endpoint='/users', object_class=User, is_sec_policy=False, is_global=True),
            None, None, '/users'
        ),
        (
            _API(name='labels', endpoint='/labels', object_class=Label, is_sec_policy=False, is_global=False),
            None, None, '/orgs/1/labels'
        ),
        (
            _API(name='rules', endpoint='/sec_rules', object_class=Rule, is_sec_policy=True, is_global=False),
            DRAFT, '/orgs/1/sec_policy/active/rule_sets/1', '/orgs/1/sec_policy/draft/rule_sets/1/sec_rules'
        ),
        (
            _API(name='rules', endpoint='/sec_rules', object_class=Rule, is_sec_policy=True, is_global=False),
            ACTIVE, '/orgs/1/sec_policy/active/rule_sets/1', '/orgs/1/sec_policy/draft/rule_sets/1/sec_rules'
        ),
        (
            _API(name='container_workload_profiles', endpoint='/container_workload_profiles', object_class=ContainerWorkloadProfile, is_sec_policy=False, is_global=False),
            ACTIVE, '/orgs/1/container_clusters/f5bef182-8c55-4219-b35b-0a50b707e434', '/orgs/1/container_clusters/f5bef182-8c55-4219-b35b-0a50b707e434/container_workload_profiles'
        )
    ]
)
def test_endpoint_building(api, policy_version, parent, expected, pce):
    endpoint = PolicyComputeEngine._PCEObjectAPI(pce, api)._build_endpoint(policy_version, parent)
    assert endpoint == expected


@pytest.fixture
def called_with():
    class Tester(object):
        def __call__(self, *args, **kwargs):
            self.args = list(args)
            return MockResponse()
    return Tester()


@pytest.mark.parametrize(
    "api_name,expected", [
        ('users', 'https://test.pce.com:443/api/v2/users'),
        ('labels', 'https://test.pce.com:443/api/v2/orgs/1/labels'),
        ('ip_lists', 'https://test.pce.com:443/api/v2/orgs/1/sec_policy/draft/ip_lists')
    ]
)
def test_internal_api_org_inclusion(api_name, expected, monkeypatch, called_with):
    monkeypatch.setattr("requests.sessions.Session.request", called_with)
    pce = PolicyComputeEngine('test.pce.com')
    api = getattr(pce, api_name)
    api.get(include_org=True)
    assert called_with.args == ['GET', expected]
    api.get_all(include_org=True)
    assert called_with.args == ['GET', expected]
    api.create({}, include_org=True)
    assert called_with.args == ['POST', expected]



@pytest.mark.parametrize(
    "api_name,href,expected", [
        ('users', '/users/1', 'https://test.pce.com:443/api/v2/users/1'),
        ('labels', '/orgs/1/labels/1', 'https://test.pce.com:443/api/v2/orgs/1/labels/1'),
        ('ip_lists', '/orgs/1/sec_policy/draft/ip_lists/1', 'https://test.pce.com:443/api/v2/orgs/1/sec_policy/draft/ip_lists/1')
    ]
)
def test_internal_api_org_inclusion_with_href(api_name, href, expected, monkeypatch, called_with):
    monkeypatch.setattr("requests.sessions.Session.request", called_with)
    pce = PolicyComputeEngine('test.pce.com')
    api = getattr(pce, api_name)
    api.get_by_reference(href, include_org=True)
    assert called_with.args == ['GET', expected]
    api.update(href, {}, include_org=True)
    assert called_with.args == ['PUT', expected]
    api.delete(href, include_org=True)
    assert called_with.args == ['DELETE', expected]


@pytest.mark.parametrize(
    "endpoint,include_org,expected", [
        ('/health', False, 'https://test.pce.com:443/api/v2/health'),
        ('labels', True, 'https://test.pce.com:443/api/v2/orgs/1/labels'),
        ('/labels', True, 'https://test.pce.com:443/api/v2/orgs/1/labels'),
        ('//labels', True, 'https://test.pce.com:443/api/v2/orgs/1/labels'),
        ('/sec_policy/active//ip_lists', True, 'https://test.pce.com:443/api/v2/orgs/1/sec_policy/active/ip_lists'),
        ('/orgs/1/workloads/ef7f0f53-2295-4416-aaaf-965146934c53', True, 'https://test.pce.com:443/api/v2/orgs/1/workloads/ef7f0f53-2295-4416-aaaf-965146934c53'),
        ('/orgs/1/workloads/ef7f0f53-2295-4416-aaaf-965146934c53', False, 'https://test.pce.com:443/api/v2/orgs/1/workloads/ef7f0f53-2295-4416-aaaf-965146934c53'),
        ('/orgs/1/sec_policy/rule_sets/1/sec_rules/1', True, 'https://test.pce.com:443/api/v2/orgs/1/sec_policy/rule_sets/1/sec_rules/1')
    ]
)
def test_url_building(endpoint, include_org, expected, pce):
    assert pce._build_url(endpoint, include_org) == expected


@pytest.mark.parametrize(
    "endpoint,expected", [
        ('/health', 'https://test.pce.com:443/api/v2/health'),
        ('/labels', 'https://test.pce.com:443/api/v2/labels'),
        ('/sec_policy/active//ip_lists', 'https://test.pce.com:443/api/v2/sec_policy/active/ip_lists'),
        ('/orgs/1/workloads/ef7f0f53-2295-4416-aaaf-965146934c53', 'https://test.pce.com:443/api/v2/orgs/1/workloads/ef7f0f53-2295-4416-aaaf-965146934c53'),
        ('/orgs/1/sec_policy/rule_sets/1/sec_rules/1', 'https://test.pce.com:443/api/v2/orgs/1/sec_policy/rule_sets/1/sec_rules/1')
    ]
)
def test_set_include_org_default(endpoint, expected):
    pce = PolicyComputeEngine('test.pce.com')
    pce.include_org = False
    response = pce.get(endpoint)
    assert response.url == expected


def test_ignore_include_org(pce):
    assert pce.check_connection(include_org=True)


@pytest.mark.parametrize(
    "api_name,endpoint,ObjectClass,is_sec_policy",
    [
        (name, api.endpoint, api.object_class, api.is_sec_policy)
        for name, api in PCE_APIS.items()
    ]
)
def test_pce_apis(api_name, endpoint, ObjectClass, is_sec_policy, pce,
        requests_mock, get_callback, post_callback, put_callback, delete_callback):
    api = getattr(pce, api_name)

    pattern = re.compile(endpoint)
    requests_mock.register_uri('GET', pattern, json=get_callback)
    requests_mock.register_uri('POST', pattern, json=post_callback)
    requests_mock.register_uri('PUT', pattern, json=put_callback)
    requests_mock.register_uri('DELETE', pattern, json=delete_callback)

    params = {}
    if 'name' in [field.name for field in fields(ObjectClass)]:
        params = {'name': 'test object'}

    obj = api.create(ObjectClass(**params))
    assert obj.href

    policy_version = DRAFT if is_sec_policy else None
    assert api.get_by_reference(obj.href) == obj

    if params:
        objs = api.get(params=params, policy_version=policy_version)
        assert objs[0] == obj

    api.update(obj.href, {'description': 'updated description'})
    obj = api.get_by_reference(obj.href)
    assert obj.description == 'updated description'

    api.delete(obj.href)
    objs = api.get(policy_version=policy_version)
    assert len(objs) == 0



# -*- coding: utf-8 -*-

"""This module provides classes related to policy rule sets.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List

from illumio.util import MutableObject, pce_api
from illumio.policyobjects import LabelSet

from .rule import Rule
from .iptablesrule import IPTablesRule


@dataclass
@pce_api('rule_sets', is_sec_policy=True)
class RuleSet(MutableObject):
    """Represents a rule set object in the PCE.

    Rule sets provide scope boundaries for security policy rules. Scopes are
    defined using application, environment, and location labels. Rules within
    the set will default to applying to workloads with these labels.

    See https://docs.illumio.com/core/21.5/Content/Guides/security-policy/create-security-policy/rulesets.htm

    Usage:
        >>> import illumio
        >>> pce = illumio.PolicyComputeEngine('pce.company.com', port=443, org_id=1)
        >>> pce.set_credentials('api_key', 'api_secret')
        >>> app_label = pce.labels.create({'key': 'app', 'value': 'A-App'})
        >>> env_label = pce.labels.create({'key': 'env', 'value': 'E-Prod'})
        >>> loc_label = pce.labels.create({'key': 'loc', 'value': 'L-AWS'})
        >>> ruleset = illumio.RuleSet(
        ...     name='RS-RINGFENCE',
        ...     scopes=[
        ...         illumio.LabelSet(
        ...             labels=[app_label, env_label, loc_label]
        ...         )
        ...     ]
        ... )
        >>> ruleset = pce.rule_sets.create(ruleset)
        >>> ruleset
        Ruleset(
            href='/orgs/1/sec_policy/draft/rule_sets/19',
            name='RS-RINGFENCE'
        )
    """
    enabled: bool = None
    scopes: List[LabelSet] = None
    rules: List[Rule] = None
    ip_tables_rules: List[IPTablesRule] = None


__all__ = [
    'RuleSet',
]



# -*- coding: utf-8 -*-

"""This module provides classes related to services and service ports.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List, Union

from illumio import IllumioException
from illumio.util import (
    JsonObject,
    MutableObject,
    ICMP_CODE_MAX,
    ICMP_TYPE_MAX,
    PORT_MAX,
    pce_api,
    convert_protocol,
    validate_int
)


@dataclass
class BaseService(JsonObject):
    port: int = None
    proto: Union[str, int] = None

    def __post_init__(self):
        if type(self.proto) is str:
            self.proto = int(self.proto) if self.proto.isnumeric() else convert_protocol(self.proto)
        super().__post_init__()

    def _validate(self):
        if self.port:
            validate_int(self.port, maximum=PORT_MAX)
        super()._validate()


@dataclass
class ServicePort(BaseService):
    """Represents a port, port range, Windows service, or traffic flow service."""
    to_port: int = None
    icmp_type: int = None
    icmp_code: int = None
    service_name: str = None
    process_name: str = None
    windows_service_name: str = None
    user_name: str = None

    def _validate(self):
        if self.to_port:
            validate_int(self.to_port, maximum=PORT_MAX)
            if self.to_port <= self.port:
                raise IllumioException("Invalid port range: to_port must be higher than port")
        if self.icmp_type:
            validate_int(self.icmp_type, maximum=ICMP_TYPE_MAX)
        if self.icmp_code:
            validate_int(self.icmp_code, maximum=ICMP_CODE_MAX)
        super()._validate()


@dataclass
@pce_api('services', is_sec_policy=True)
class Service(MutableObject):
    """Represents a service in the PCE.

    A service can be port-based or process-based (Windows services).

    Each service contains one or more objects defining the port, protocol, and/or
    process name used by an application running on a workload.

    Service objects are used to write rules or enforcement boundaries to allow
    or deny traffic on its defined ports and processes for workloads in the network.

    See https://docs.illumio.com/core/21.5/Content/Guides/security-policy/security-policy-objects/services.htm

    Usage:
        >>> import illumio
        >>> pce = illumio.PolicyComputeEngine('pce.company.com', port=443, org_id=1)
        >>> pce.set_credentials('api_key', 'api_secret')
        >>> service = illumio.Service(
        ...     name='S-HTTP',
        ...     service_ports=[
        ...         illumio.ServicePort(port=80, proto='tcp'),
        ...         illumio.ServicePort(port=443, proto='tcp')
        ...     ]
        ... )
        >>> service = pce.services.create(service)
        >>> service
        Service(
            href='/orgs/1/sec_policy/draft/services/15',
            name='S-HTTP',
            service_ports=[
                ServicePort(
                    port=80,
                    proto=6,
                    ...
                ),
                ...
            ],
            ...
        )
    """
    process_name: str = None
    service_ports: List[ServicePort] = None
    windows_services: List[ServicePort] = None
    windows_egress_services: List[ServicePort] = None


__all__ = [
    'BaseService',
    'ServicePort',
    'Service',
]



import json
import os
import re
from typing import List

import pytest

from illumio.workloads import VEN

VENS = os.path.join(pytest.DATA_DIR, 'vens.json')


@pytest.fixture(scope='module')
def vens() -> List[dict]:
    with open(VENS, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(autouse=True)
def vens_mock(pce_object_mock, vens):
    pce_object_mock.add_mock_objects(vens)


@pytest.fixture(autouse=True)
def mock_requests(requests_mock, get_callback):
    pattern = re.compile('/vens')
    requests_mock.register_uri('GET', pattern, json=get_callback)


@pytest.fixture()
def mock_ven(pce) -> VEN:
    yield pce.vens.get_by_reference("/orgs/1/vens/ec38510d-e4fd-41a1-a6ba-8b0bb4ce9ae9")


def test_get_by_hostname(pce, mock_ven):
    ven = pce.vens.get(params={'hostname': 'WIN-CK9JH7R07NB'})[0]
    assert ven == mock_ven



# -*- coding: utf-8 -*-

"""The illumio library provides a simple interface for interacting with PCE APIs.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from .events import *
from .exceptions import *
from .secpolicy import *
from .util import *
from .accessmanagement import *
from .policyobjects import *
from .infrastructure import *
from .vulnerabilities import *
from .workloads import *
from .rules import *
from .explorer import *
from .pce import *

from types import ModuleType

# avoid name conflicts with package modules when using
# `from illumio import *` by excluding them here
__all__ = [
    export for export, o in globals().items()
        if not (export.startswith('_') or isinstance(o, ModuleType))
]



# -*- coding: utf-8 -*-

"""This module provides classes related to pairing profile objects.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List, Union

from illumio import IllumioException
from illumio.util import (
    Reference,
    MutableObject,
    EnforcementMode,
    VisibilityLevel,
    pce_api
)


@dataclass
@pce_api('pairing_profiles')
class PairingProfile(MutableObject):
    """Represents a pairing profile in the PCE.

    Pairing profiles are used to configure VEN defaults and generate keys
    for VEN pairing  on workloads that will be managed by the PCE.

    See https://docs.illumio.com/asp/20.1/Content/Guides/rest-api/workloads/pairing-profiles-and-pairing-keys.htm

    Usage:
        >>> import illumio
        >>> pce = illumio.PolicyComputeEngine('pce.company.com', port=443, org_id=1)
        >>> pce.set_credentials('api_key', 'api_secret')
        >>> pairing_profile = illumio.PairingProfile(
        ...     name='PP-DATABASE-VENS',
        ...     enabled=True,
        ...     enforcement_mode='visibility_only',
        ...     visibility_level='flows_summary'
        ... )
        >>> pairing_profile = pce.pairing_profiles.create(pairing_profile)
        >>> pairing_profile
        PairingProfile(
            href='/orgs/1/pairing_profiles/19',
            name='PP-DATABASE-VENS',
            enabled=True,
            enforcement_mode='visibility_only',
            visibility_level='flows_summary',
            ...
        )
    """
    enabled: bool = None
    agent_software_release: str = None
    enforcement_mode: str = None
    enforcement_mode_lock: bool = None
    visibility_level: str = None
    visibility_level_lock: bool = None
    allowed_uses_per_key: Union[str, int] = None
    key_lifespan: Union[str, int] = None

    labels: List[Reference] = None
    role_label_lock: bool = None
    app_label_lock: bool = None
    env_label_lock: bool = None
    loc_label_lock: bool = None

    total_use_count: int = None
    is_default: bool = None
    last_pairing_at: str = None

    def _validate(self):
        if self.enforcement_mode and not self.enforcement_mode in EnforcementMode:
            raise IllumioException("Invalid enforcement_mode: {}".format(self.enforcement_mode))
        if self.visibility_level and not self.visibility_level in VisibilityLevel:
            raise IllumioException("Invalid visibility_level: {}".format(self.visibility_level))
        super()._validate()

    def _encode_field(self, field):
        """Encodes PairingProfile fields to JSON types.

        The allowed_uses_per_key and key_lifespan parameters have strict schema
        contraints in the PCE - values must either be 'unlimited' or an integer
        which we enforce here. All other fields use the default encoding method.
        """
        if field.name in ['allowed_uses_per_key', 'key_lifespan']:
            value = getattr(self, field.name)
            if isinstance(value, str) and value != 'unlimited':
                if not value.isnumeric():
                    raise IllumioException("Invalid '{}' value: must be 'unlimited' or int".format(field.name))
                value = int(value)
            return value
        return super()._encode_field(field)


__all__ = [
    'PairingProfile',
]



import json
import os
import re
from typing import List

import pytest

from illumio import Event

EVENTS = os.path.join(pytest.DATA_DIR, 'events.json')


@pytest.fixture(scope='module')
def events() -> List[dict]:
    with open(EVENTS, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(autouse=True)
def events_mock(pce_object_mock, events):
    pce_object_mock.add_mock_objects(events)


@pytest.fixture(autouse=True)
def mock_requests(requests_mock, get_callback):
    pattern = re.compile('/events')
    requests_mock.register_uri('GET', pattern, json=get_callback)


@pytest.fixture()
def mock_event(pce) -> Event:
    yield pce.events.get_by_reference("/orgs/1/events/a066c6ec-4d4a-4c51-bf46-9a20add3bcac")


def test_get_by_event_type(pce, mock_event):
    event = pce.events.get(params={'event_type': 'user.login'})[0]
    assert event == mock_event



# -*- coding: utf-8 -*-

"""This module provides classes related to Virtual Enfocement Node objects.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass, field
from typing import List

from illumio import IllumioException, NotificationEvent
from illumio.infrastructure import ContainerCluster
from illumio.util import (
    JsonObject,
    Reference,
    MutableObject,
    LinkState,
    VENType,
    VisibilityLevel,
    pce_api,
)


@dataclass
class Interface(Reference):
    name: str = None
    link_state: str = None
    address: str = None
    cidr_block: int = None
    default_gateway_address: str = None
    network: Reference = None
    network_detection_mode: str = None
    friendly_name: str = None
    loopback: bool = None

    def _validate(self):
        if self.link_state and not self.link_state in LinkState:
            raise IllumioException("Invalid link_state: {}".format(self.link_state))
        super()._validate()


@dataclass
class AgentConfig(JsonObject):
    mode: str = None
    log_traffic: bool = None
    security_policy_update_mode: str = None
    visibility_level: str = None

    def _validate(self):
        if self.visibility_level and not self.visibility_level in VisibilityLevel:
            raise IllumioException("Invalid visibility_level: {}".format(self.visibility_level))
        super()._validate()


@dataclass
class SecureConnect(JsonObject):
    matching_issuer_name: str = None


@dataclass
class AgentHealthErrors(JsonObject):
    errors: List[str] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)


@dataclass
class AgentHealth(JsonObject):
    type: str = None
    severity: str = None
    audit_event: str = None


@dataclass
class AgentStatus(JsonObject):
    status: str = None
    uid: str = None
    instance_id: str = None
    last_heartbeat_on: str = None
    uptime_seconds: int = None
    agent_version: str = None
    managed_since: str = None
    fw_config_current: bool = None
    firewall_rule_count: int = None
    security_policy_refresh_at: str = None
    security_policy_applied_at: str = None
    security_policy_received_at: str = None
    agent_health_errors: AgentHealthErrors = None
    agent_health: List[AgentHealth] = None
    security_policy_sync_state: str = None


@dataclass
class VENAgent(MutableObject):
    config: AgentConfig = None
    secure_connect: SecureConnect = None
    status: AgentStatus = None
    unpair_allowed: bool = None
    active_pce_fqdn: str = None
    target_pce_fqdn: str = None
    type: str = None


@dataclass
class SecureConnect(JsonObject):
    matching_issuer_name: str = None


@dataclass
class VENCondition(JsonObject):
    first_reported_timestamp: str = None
    latest_event: NotificationEvent = None


@dataclass
@pce_api("vens")
class VEN(MutableObject):
    """Represents a Virtual Enforcement Node object in the PCE.

    The VEN object represents the agent on a workload that has been paired with
    the PCE.

    **NOTE:** VENs are read-only via the PCE API.

    See https://docs.illumio.com/core/21.5/Content/Guides/rest-api/workloads/ven-operations.htm

    Usage:
        >>> import illumio
        >>> pce = illumio.PolicyComputeEngine('pce.company.com', port=443, org_id=1)
        >>> pce.set_credentials('api_key', 'api_secret')
        >>> vens = pce.vens.get()
        >>> vens
        [
            VEN(
                href="/orgs/1/vens/5fe88f13-d22c-427e-be47-b9c3369e1e32",
                hostname="web01.lab.company.com",
                status="active",
                os_id="centos-x86_64-8.0",
                os_detail="5.18.11-200.fc36.x86_64 #1 SMP PREEMPT_DYNAMIC Tue Jul 12 22:52:35 UTC 2022 (CentOS Linux release 8.5.2111)",
                os_platform="linux",
                version="21.2.5-8017",
                activation_type="pairing_key",
                labels=[
                    Reference(href="/orgs/1/labels/14"),
                    ...
                ],
                interfaces=[
                    Interface(
                        href="/orgs/1/workloads/5fe88f13-d22c-427e-be47-b9c3369e1e32/interfaces/eth0",
                        name="eth0",
                        address="10.88.0.3",
                        cidr_block=16,
                        default_gateway_address="10.88.0.1",
                        network=Reference(
                            href="/orgs/1/networks/40995e57-5e83-4fc5-9492-5b0e6df32a68"
                        ),
                        network_detection_mode="single_private_brn",
                        loopback=False,
                    ),
                    ...
                ],
                workloads=[
                    Reference(href="/orgs/1/workloads/5fe88f13-d22c-427e-be47-b9c3369e1e32")
                ],
                last_heartbeat_at="2022-08-01T17:14:08.830Z",
                last_goodbye_at=None,
                unpair_allowed=True,
                conditions=[
                    VENCondition(
                        first_reported_timestamp="2022-08-01T17:31:40.669Z",
                        latest_event=Event(
                            href="/orgs/1/events/0d2cec5f-5025-4e62-b867-b6e607397105",
                            notification_type="agent.missed_heartbeats",
                            severity="warning",
                            info={
                                ...
                            },
                            timestamp="2022-08-01T17:31:40.669Z",
                        ),
                    ),
                    ...
                ],
            ),
            ...
        ]
    """

    hostname: str = None
    status: str = None
    uid: str = None
    os_id: str = None
    os_detail: str = None
    os_platform: str = None
    version: str = None
    status: str = None
    activation_type: str = None
    active_pce_fqdn: str = None
    target_pce_fqdn: str = None
    labels: List[Reference] = None
    interfaces: List[Interface] = None
    workloads: List[Reference] = None
    container_cluster: ContainerCluster = None
    secure_connect: SecureConnect = None
    last_heartbeat_at: str = None
    last_goodbye_at: str = None
    unpair_allowed: bool = None
    conditions: List[VENCondition] = None
    ven_type: str = None

    def _validate(self):
        if self.ven_type and self.ven_type not in VENType:
            raise IllumioException("Invalid ven_type: {}".format(self.ven_type))
        super()._validate()


__all__ = [
    "AgentConfig",
    "SecureConnect",
    "AgentHealthErrors",
    "AgentHealth",
    "AgentStatus",
    "VENAgent",
    "VEN",
]



import json
import os
import re
from typing import List

import pytest

from illumio.infrastructure import ContainerWorkloadProfile
from illumio.util import EnforcementMode

CONTAINER_WORKLOAD_PROFILES = os.path.join(pytest.DATA_DIR, 'container_workload_profiles.json')
CONTAINER_CLUSTER_HREF = '/orgs/1/container_clusters/f5bef182-8c55-4219-b35b-0a50b707e434'


@pytest.fixture(scope='module')
def container_workload_profiles() -> List[dict]:
    with open(CONTAINER_WORKLOAD_PROFILES, 'r') as f:
        yield json.loads(f.read())


@pytest.fixture(autouse=True)
def container_clusters_mock(pce_object_mock, container_workload_profiles):
    pce_object_mock.add_mock_objects(container_workload_profiles)


@pytest.fixture(autouse=True)
def mock_requests(requests_mock, get_callback, post_callback, put_callback, delete_callback):
    pattern = re.compile('/container_workload_profiles')
    requests_mock.register_uri('GET', pattern, json=get_callback)
    requests_mock.register_uri('POST', pattern, json=post_callback)
    requests_mock.register_uri('PUT', pattern, json=put_callback)
    requests_mock.register_uri('DELETE', pattern, json=delete_callback)


def test_get_workload_profiles_by_cluster(pce):
    workload_profiles = pce.container_workload_profiles.get(parent=CONTAINER_CLUSTER_HREF)
    assert len(workload_profiles) == 9


def test_get_managed_workload_profiles(pce):
    workload_profiles = pce.container_workload_profiles.get(params={'managed': True}, parent=CONTAINER_CLUSTER_HREF)
    assert len(workload_profiles) == 2


def test_add_workload_profile_to_cluster(pce):
    workload_profile = ContainerWorkloadProfile(namespace='illumio-system')
    workload_profile = pce.container_workload_profiles.create(workload_profile, parent=CONTAINER_CLUSTER_HREF)
    workload_profiles = pce.container_workload_profiles.get(parent=CONTAINER_CLUSTER_HREF)
    assert workload_profile.href in [profile.href for profile in workload_profiles]


def test_update_workload_profile(pce):
    workload_profile = pce.container_workload_profiles.get(params={'namespace': 'illumio-system', 'max_results': 1})[0]
    pce.container_workload_profiles.update(workload_profile.href, {'enforcement_mode': EnforcementMode.FULL})
    updated_workload_profile = pce.container_workload_profiles.get_by_reference(workload_profile.href)
    assert updated_workload_profile.enforcement_mode == EnforcementMode.FULL.value




# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

if sys.version_info < (3, 8):
    from importlib_metadata import version
else:
    from importlib.metadata import version

sys.path.insert(0, os.path.abspath("../.."))

from illumio import PCE_APIS, BULK_CHANGE_LIMIT

# -- Project information -----------------------------------------------------

project = 'illumio'
copyright = '2022, Illumio'
author = 'Illumio'

# The full version, including alpha/beta/rc tags
release = version('illumio')

# Simple x.y.z version number
version = '.'.join(release.split('.')[:3])

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.napoleon',
    'sphinx.ext.autodoc',
    'enum_tools.autoenum',
    'sphinx.ext.viewcode'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    'tests'
]

add_module_names = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'furo'
html_theme_options = {
    'navigation_with_keys': True,
    'sidebar_hide_name': True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Include custom CSS styles.
html_css_files = ['custom.css']

# Use the Illumio logo in place of the library name.
html_logo = 'illumio_logo.svg'

# Hide the Made with Sphinx footer.
html_show_sphinx = False

# -- Autodoc Options ---------------------------------------------------------

# Exclude documentation for __new__ by default.
autodoc_default_options = {
    'exclude-members': '__init__, __new__'
}

# Override the default alphabetical ordering for member variables and methods,
# instead display them in source definition order.
autodoc_member_order = 'bysource'

# Show type hints in the description rather than the signature.
autodoc_typehints = 'description'

# Don't show class signature with the class name.
autodoc_class_signature = 'separated'

# Don't inherit docstrings from parent classes.
autodoc_inherit_docstrings = False

# -- Custom definitions ------------------------------------------------------

# Construct the dynamic API list to embed into the API documentation page.
apis_docstring = ''
for name, api in sorted(PCE_APIS.items()):
    classpath = '{}.{}'.format(
        api.object_class.__module__,
        api.object_class.__qualname__
    )
    apis_docstring += """
    :class:`{} <{}>` |br|""".format(name, classpath)

# Adds the rst string to the bottom of every page. We use this to define
# common and dynamic substitutions.
rst_epilog = """
.. |br| raw:: html

    <br/>
.. |APIList| replace:: {}
.. |BulkChangeLimit| replace:: {}
""".format(apis_docstring, BULK_CHANGE_LIMIT)



import pytest

from illumio.infrastructure import ContainerCluster

from helpers import random_string


@pytest.fixture
def container_cluster(pce, session_identifier):
    identifier = random_string()
    container_cluster = pce.container_clusters.create(
        ContainerCluster(
            name='{}-CC-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests'
        )
    )
    yield container_cluster
    pce.container_clusters.delete(container_cluster.href)

__all__ = ['container_cluster']



# -*- coding: utf-8 -*-

"""This module provides the core of the class model used in the client.

The base JsonObject class handles basic encoding and decoding to translate
responses from the PCE REST API into python objects. The implementation
leverages dataclasses to simplify the logic for these operations.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
import copy
import json
from abc import ABC
from dataclasses import Field, dataclass, fields
from inspect import signature, isclass
from typing import List, Any, Union

from illumio.exceptions import IllumioException

from .constants import IllumioEnumMeta
from .functions import ignore_empty_keys, isunion, islist

_default = json.JSONEncoder()  # fall back to the default encoder for non-Illumio API objects


class IllumioEncoder(json.JSONEncoder):
    """Convenience class for encoding JsonObjects.

    >>> json.dumps(flow, cls=IllumioEncoder, indent=4)
    """
    def default(self, o: Any) -> Any:
        return getattr(o.__class__, "to_json", _default.default)(o)


@dataclass
class JsonObject(ABC):
    """Base dataclass for all derived PCE objects.

    Provides custom encoding, decoding, and type validation to and from JSON.
    """

    def __post_init__(self):
        self._validate()

    def _validate(self):
        """Validates all dataclass fields against their indicated types."""
        for field in fields(self):
            value = getattr(self, field.name)
            if not self._validate_field(field.type, value):
                raise AttributeError("Invalid value for {}: {}. Must be of type {}".format(field.name, value, field.type))

    def _validate_field(self, expected_type, value) -> bool:
        if value is None:
            return True
        elif expected_type is object:
            return True
        elif type(value) == expected_type:
            return True
        elif isunion(expected_type):
            return any(self._validate_field(type_, value) for type_ in expected_type.__args__)
        elif isinstance(type(value), IllumioEnumMeta):
            # validate enum values if they are passed as enum consts
            return self._validate_field(expected_type, value.value)
        elif isclass(expected_type) and issubclass(expected_type, JsonObject):
            # if the object is already decoded, determine whether it's
            # a subtype of what the field expects
            if isinstance(value, JsonObject):
                return isinstance(value, expected_type)
            try:
                # XXX: otherwise, expand objects to run their own validation.
                #   this is *slow*, but needed to validate deeply nested types
                expected_type.from_json(value)
            except:
                return False
            return True
        elif islist(expected_type) and isinstance(value, list):
            if not value:
                return True  # empty lists are always valid
            expected_type = expected_type.__args__[0]
            return all(self._validate_field(expected_type, o) for o in value)
        return False

    def to_json(self) -> Any:
        """Converts the object to a JSON-compatible copy of itself.

        Objects are converted to dictionaries recursively.

        Returns:
            Any: the converted JSON-compatible object
        """
        return deep_encode(self)

    def _encode(self) -> Any:
        result = []
        for f in fields(self):
            result.append((f.name, self._encode_field(f)))
        return ignore_empty_keys(result)

    def _encode_field(self, field: Field) -> Any:
        value = flatten_ref(field.type, getattr(self, field.name))
        value = resolve_enum(value)
        return deep_encode(value)

    @classmethod
    def from_json(cls, data: Any) -> 'JsonObject':
        """
        Given a JSON object or dictionary, decodes the data as an object
        of the calling JsonObject subtype. Accepts arbitrary key/value pairs
        as a form of forwards-compatibility.

        Classes can optionally extend decoding with the _decode_complex_types
        function for complex-type members.

        Based in part on https://stackoverflow.com/a/55101438
        """
        data = json.loads(data) if type(data) is str else data
        cls_fields = {field for field in signature(cls).parameters}

        defined_params, undefined_params = {}, {}
        for k, v in data.items():
            if k in cls_fields:
                defined_params[k] = v
            else:
                undefined_params[k] = v

        o = cls(**defined_params)

        for k, v in undefined_params.items():
            setattr(o, k, v)

        o._decode_complex_types()
        return o

    def _decode_complex_types(self) -> None:
        for field in fields(self):
            value = getattr(self, field.name)
            decoded_value = self._decode_field(field.type, value)
            setattr(self, field.name, decoded_value)

    def _decode_field(self, type_, value) -> Any:
        if value is None:
            return None
        if isinstance(value, JsonObject):
            # if the value has already been decoded, return it
            return value
        elif isclass(type_) and issubclass(type_, JsonObject):
            return type_.from_json(value)
        elif isinstance(value, list):
            # if the value is a list, expect the field type to be List[T]
            type_ = type_.__args__[0]
            return list(self._decode_field(type_, o) for o in value)
        return value


def flatten_ref(type_, value):
    """Replaces Reference subclasses with a simplified Reference object.

    This allows clients to pass a Reference subclass instance without
    breaking the encoded object schema for API calls.
    """
    if value is None:
        return None
    if type_ is Reference:
        if isinstance(value, Reference):
            return Reference(value.href)
    elif islist(type_):
        if type_.__args__[0] is Reference:
            ref_list = []
            for ref in value:
                if isinstance(ref, Reference):
                    ref_list.append(Reference(href=ref.href))
                else:
                    ref_list.append(ref)
            return ref_list
    elif isunion(type_):
        if Reference in type_.__args__:
            if isinstance(value, Reference):
                return Reference(value.href)
    return value


def resolve_enum(value):
    """Replaces IllumioEnumMeta subtypes with their internal value.

    This allows clients to pass enums directly as attribute values.

    For example:

    >>> Workload(..., enforcement_mode=EnforcementMode.SELECTIVE)

    will be converted to

    >>> Workload(..., enforcement_mode='selective')
    """
    if value is None:
        return None
    if isinstance(type(value), IllumioEnumMeta):
        return value.value
    return value


def deep_encode(o: Any) -> Any:
    """
    Recursively encode members of the given object and return a JSON-compatible
    copy. Children of the JsonObject superclass can optionally implement _encode
    or _encode_field to provide a customized encoding response, otherwise the
    JsonObject defaults are called to remove null-value pairs.

    Functionally similar to the dataclasses asdict method, but with the necessary
    adjustment of calling an optional custom encoding function for types that
    don't strictly mirror their dataclass field pairs when encoded.
    """
    if isinstance(o, JsonObject):
        return o._encode()
    elif isinstance(o, (list, tuple)):
        return type(o)(deep_encode(o) for o in o)
    elif isinstance(o, dict):
        return type(o)((deep_encode(k), deep_encode(v)) for k, v in o.items())
    else:
        return copy.deepcopy(o)


@dataclass
class Reference(JsonObject):
    """Simplest PCE object type, containing only an HREF.

    Used in most API schema to refer to other PCE objects.

    Args:
        href (str, optional): PCE object HREF.
    """
    href: str = None


def href_from(reference: Union[Reference, dict, str]):
    """Attempts to parse HREF value from a provided source.

    Args:
        reference (Union[Reference, dict, str]): source reference. If a string
            value is passed, it is returned unchanged. The ``href`` field
            is returned from a Reference object or subclass instance, and the
            ``'href'`` key is returned from a provided dictionary.

    Raises:
        IllumioException: if an invalid reference type is provided, or if the
            href value is null or falsy.
    """
    if isinstance(reference, Reference):
        if reference.href:
            return reference.href
    elif type(reference) is dict:
        if 'href' in reference and reference['href']:
            return reference['href']
    elif type(reference) is str:
        return reference
    raise IllumioException('Failed to extract HREF from value: {}'.format(reference))


@dataclass
class IllumioObject(Reference):
    """Base class for most PCE objects.

    Args:
        name (str, optional): object name.
        description (str, optional): object description.
        external_data_set (str, optional): unique namespace identifier for an
            external source creating PCE objects. If set,
            ``external_data_reference`` must also be provided.
        external_data_reference (str, optional): unique identifier within the
            external_data_set. If set, ``external_data_set`` must also be
            provided. ``external_data_set`` + ``external_data_reference``
            must be globally unique.
        caps (List[str], optional): defines the requesting user's
            capabilities/permissions on the object.
    """
    name: str = None
    description: str = None
    external_data_set: str = None
    external_data_reference: str = None
    caps: List[str] = None


@dataclass
class MutableObject(IllumioObject):
    """Base class for PCE objects that can be updated/deleted."""
    created_at: str = None
    updated_at: str = None
    deleted_at: str = None
    update_type: str = None
    delete_type: str = None
    created_by: Reference = None
    updated_by: Reference = None
    deleted_by: Reference = None


@dataclass
class ImmutableObject(IllumioObject):
    """Base class for PCE objects that cannot be updated/deleted."""
    created_at: str = None
    created_by: Reference = None


@dataclass
class Error(JsonObject):
    """Wrapper class for error status/message.

    Args:
        token (str, optional): error status.
        message (str, optional): error message.
    """
    token: str = None
    message: str = None


__all__ = [
    'IllumioEncoder',
    'JsonObject',
    'Reference',
    'IllumioObject',
    'MutableObject',
    'ImmutableObject',
    'Error',
    'href_from',
]



import pytest

from illumio.policyobjects import Label

from helpers import random_string


@pytest.fixture
def role_label(pce, session_identifier):
    identifier = random_string()
    label = pce.labels.create(
        Label(
            key='role',
            value='{}-R-{}'.format(session_identifier, identifier),
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield label
    pce.labels.delete(label.href)


@pytest.fixture
def app_label(pce, session_identifier):
    identifier = random_string()
    label = pce.labels.create(
        Label(
            key='app',
            value='{}-A-{}'.format(session_identifier, identifier),
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield label
    pce.labels.delete(label.href)


@pytest.fixture
def env_label(pce, session_identifier):
    identifier = random_string()
    label = pce.labels.create(
        Label(
            key='env',
            value='{}-E-{}'.format(session_identifier, identifier),
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield label
    pce.labels.delete(label.href)


@pytest.fixture
def loc_label(pce, session_identifier):
    identifier = random_string()
    label = pce.labels.create(
        Label(
            key='loc',
            value='{}-L-{}'.format(session_identifier, identifier),
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield label
    pce.labels.delete(label.href)

__all__ = [
    'role_label',
    'app_label',
    'env_label',
    'loc_label'
]



"""This module defines teardown functions for integration tests.


"""
from illumio import PolicyComputeEngine, IllumioException


class Sweeper:
    def __init__(self, pce: PolicyComputeEngine, session_identifier) -> None:
        self.pce = pce
        self.session_identifier = session_identifier

    def _sweep(self, api, lookup='external_data_set'):
        try:
            objects = api.get(params={lookup: self.session_identifier})
            for o in objects:
                api.delete(o.href)
        except IllumioException:
            pass

    def _sweep_container_clusters(self):
        self._sweep(self.pce.container_clusters, lookup='name')

    def _sweep_enforcement_boundaries(self):
        self._sweep(self.pce.enforcement_boundaries, lookup='name')

    def _sweep_ip_lists(self):
        self._sweep(self.pce.ip_lists)

    def _sweep_label_groups(self):
        self._sweep(self.pce.label_groups)

    def _sweep_labels(self):
        self._sweep(self.pce.labels)

    def _sweep_pairing_profiles(self):
        self._sweep(self.pce.pairing_profiles)

    def _sweep_rule_sets(self):
        self._sweep(self.pce.rule_sets)

    def _sweep_services(self):
        self._sweep(self.pce.services)

    def _sweep_virtual_services(self):
        self._sweep(self.pce.virtual_services)

    def _sweep_workloads(self):
        self._sweep(self.pce.workloads)

    def sweep(self):
        [getattr(self, fn)() for fn in dir(self) if fn.startswith('_sweep_')]



# -*- coding: utf-8 -*-

"""This module provides classes related to virtual services.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List

from illumio import IllumioException
from illumio.util import (
    IllumioObject,
    Reference,
    MutableObject,
    ApplyTo,
    PORT_MAX,
    pce_api,
    validate_int
)

from .service import BaseService, ServicePort


@dataclass
class ServiceAddress(BaseService):
    """Defines a Service Address record for use in a Virtual Service address pool.

    A ServiceAddress can be defined with either an IP address or FQDN.

    If an IP is provided, one of a port number or network object HREF
    must be provided as well.

    If an FQDN is given it's sufficient by itself, but description or
    port values can optionally be provided.
    """
    ip: str = None
    fqdn: str = None
    network: Reference = None
    description: str = None


@dataclass
@pce_api('virtual_services', is_sec_policy=True)
class VirtualService(MutableObject):
    """Represents a virtual service object in the PCE.

    Virtual services provide an abstraction for a service that can be bound to
    one or more workloads. This can be useful when a workload is running multiple
    services, or multiple instances of the same service for different apps.

    See https://docs.illumio.com/core/21.5/Content/Guides/security-policy/security-policy-objects/virtual-services.htm

    Usage:
        >>> import illumio
        >>> pce = illumio.PolicyComputeEngine('pce.company.com', port=443, org_id=1)
        >>> pce.set_credentials('api_key', 'api_secret')
        >>> role_label = pce.labels.create({'key': 'role', 'value': 'R-Tomcat'})
        >>> app_label = pce.labels.create({'key': 'app', 'value': 'A-App'})
        >>> env_label = pce.labels.create({'key': 'env', 'value': 'E-Dev'})
        >>> loc_label = pce.labels.create({'key': 'loc', 'value': 'L-NYC'})
        >>> tomcat_svc = pce.services.create(illumio.Service(
        ...     name='S-Tomcat',
        ...     service_ports=[
        ...         {
        ...             'port': 80,
        ...             'proto': illumio.convert_protocol('tcp')
        ...         },
        ...         {
        ...             'port': 443,
        ...             'proto': illumio.convert_protocol('tcp')
        ...         }
        ...     ]
        ... )
        >>> virtual_service = illumio.VirtualService(
        ...     name='VS-Tomcat',
        ...     apply_to=illumio.ApplyTo.HOST_ONLY,
        ...     service=tomcat_svc,
        ...     labels=[role_label, app_label, env_label, loc_label]
        ... )
        >>> virtual_service = pce.virtual_services.create(virtual_service)
        >>> virtual_service
        VirtualService(
        {
            "href": "/orgs/1/sec_policy/draft/virtual_services/14d7ff69-2fa4-458b-a299-e3f11ffa9b01",
            "created_at": "2021-10-05T12:34:56.789Z",
            "updated_at": "2021-10-05T12:34:56.789Z",
            "deleted_at": null,
            "created_by": {
                "href": "/users/1"
            },
            "updated_by": {
                "href": "/users/1"
            },
            "deleted_by": null,
            "update_type": "create",
            "name": "VS-INTERNAL",
            "description": null,
            "pce_fqdn": null,
            "service_ports": [
                {
                    "port": 1234,
                    "proto": 6
                },
                {
                    "port": 80,
                    "proto": 17,
                    "to_port": 443
                }
            ],
            "service_addresses": [
                {
                    "ip": "1.1.1.1",
                    "port": 101
                },
                {
                    "description": "test description",
                    "fqdn": "*.illumio.com"
                }
            ],
            "labels": [
                {
                    "href": "/orgs/1/labels/1",
                    "key": "role",
                    "value": "R-WEB"
                },
                {
                    "href": "/orgs/1/labels/2",
                    "key": "app",
                    "value": "A-HRM"
                },
                {
                    "href": "/orgs/1/labels/3",
                    "key": "env",
                    "value": "E-STAGE"
                },
                {
                    "href": "/orgs/1/labels/4",
                    "key": "loc",
                    "value": "L-TOR"
                }
            ],
            "ip_overrides": [
                "1.2.3.4"
            ],
            "apply_to": "host_only",
            "caps": [
                "write",
                "provision",
                "delete"
            ]
        }
        )

    Raises:
        IllumioException: if an invalid apply_to value is provided.
    """
    apply_to: str = None
    pce_fqdn: str = None
    service: Reference = None
    service_ports: List[ServicePort] = None
    service_addresses: List[ServiceAddress] = None
    ip_overrides: List[str] = None
    labels: List[Reference] = None

    def _validate(self):
        if self.apply_to and self.apply_to not in ApplyTo:
            raise IllumioException("Invalid 'apply_to' value: {}".format(self.apply_to))
        super()._validate()


@dataclass
class PortOverride(BaseService):
    """Represents a service binding port override.

    The ``proto`` field can be provided as a string, e.g. 'tcp', and it will be
    converted to its IANA integer equivalent.

    Args:
        port (int, optional): virtual service port to override. If the port is
            part of a range in the service, this must be the starting port in
            the range.
        proto (Union[str, int], optional): IANA protocol name or number for the
            service protocol.
        new_port (int, optional): override port or starting port of an override
            range.
        new_to_port (int, optional): ending port in an override range.

    Raises:
        IllumioException: if an invalid port value or port range is given.
    """
    new_port: int = None
    new_to_port: int = None

    def _validate(self):
        if self.new_port:
            validate_int(self.new_port, maximum=PORT_MAX)
        if self.new_to_port:
            validate_int(self.new_to_port, maximum=PORT_MAX)
            if self.new_to_port <= self.new_port:
                raise IllumioException("Invalid port range: new_to_port must be higher than new_port")
        super()._validate()


@dataclass
@pce_api('service_bindings')
class ServiceBinding(IllumioObject):
    """Represents a service binding between a virtual service and a workload.

    The ``/service_bindings`` POST method accepts a list of bindings and returns
    a list containing either the binding HREF (if the workload was successfully
    bound to the virtual service) or an error token and message.

    **NOTE:** Service bindings can only be created using an active virtual service.

    See https://docs.illumio.com/core/21.5/Content/Guides/security-policy/security-policy-objects/virtual-services.htm#BindaVirtualServicetoaWorkload

    Usage:
        >>> import illumio
        >>> workload = pce.workloads.create(illumio.Workload(
        ...     name='haproxy0.company.com',
        ...     hostname='haproxy0.company.com',
        ...     public_ip='10.0.14.201'
        ... ))
        >>> virtual_service = pce.virtual_services.create(illumio.VirtualService(
        ...     name='VS-HAProxy',
        ...     apply_to=illumio.ApplyTo.HOST_ONLY,
        ...     service_ports=[{'port': 443, 'proto': convert_protocol('tcp')}]
        ... ))
        >>> policy_version = pce.provision_policy_changes(
        ...     change_description="Create HAProxy virtual service",
        ...     hrefs=[virtual_service.href]
        ... )
        >>> virtual_service.href = convert_draft_href_to_active(virtual_service.href)
        >>> service_binding = illumio.ServiceBinding(
        ...     virtual_service=virtual_service,
        ...     workload=workload,
        ...     port_overrides=[illumio.PortOverride(port=443, new_port=8443, proto='tcp')]
        ... )
        >>> bindings = pce.service_bindings.create([service_binding])
        {
            'service_bindings': [
                ServiceBinding(href='/orgs/1/service_bindings/bc98cf25-8f24-4989-853c-578fa14108cf')
            ],
            'errors': []
        }
    """
    virtual_service: Reference = None
    workload: Reference = None
    port_overrides: List[PortOverride] = None


__all__ = [
    'ServiceAddress',
    'VirtualService',
    'PortOverride',
    'ServiceBinding',
]



import pytest

from illumio.policyobjects import LabelSet
from illumio.rules import RuleSet

from helpers import random_string


@pytest.fixture
def rule_set(pce, session_identifier, app_label, env_label, loc_label):
    identifier = random_string()
    rule_set = pce.rule_sets.create(
        RuleSet(
            name='{}-RS-{}'.format(session_identifier, identifier),
            description='Created by illumio python library integration tests',
            enabled=True,
            scopes=[LabelSet(labels=[app_label, env_label, loc_label])],
            external_data_set=session_identifier,
            external_data_reference=identifier
        )
    )
    yield rule_set
    pce.rule_sets.delete(rule_set.href)

__all__ = ['rule_set']



# -*- coding: utf-8 -*-

"""This module is a stub for classes related to vulnerability exposure scores.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass

from illumio.util import MutableObject


@dataclass
class Vulnerability(MutableObject):
    score: int = None


__all__ = [
    'Vulnerability',
]



# -*- coding: utf-8 -*-

"""This module provides common exception classes used throughout the application.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List

from .exceptions import IllumioException
from .util import (
    Reference,
    Error,
    ChangeType,
    EventStatus,
    EventSeverity,
    pce_api,
)


@dataclass
class BaseEvent(Reference):
    event: str = None
    event_type: str = None
    uuid: str = None
    timestamp: str = None
    pce_fqdn: str = None
    severity: str = None
    status: str = None
    created_by: dict = None
    info: dict = None

    def _validate(self):
        if self.severity and not self.severity in EventSeverity:
            raise IllumioException("Invalid event severity: {}".format(self.severity))
        if self.status and not self.status in EventStatus:
            raise IllumioException("Invalid event status: {}".format(self.status))
        super()._validate()


@dataclass
class NotificationEvent(BaseEvent):
    notification_type: str = None


@dataclass
class ResourceEvent(BaseEvent):
    version: str = None
    org_id: int = None
    resource: dict = None
    changes: dict = None
    change_type: str = None

    def _validate(self):
        if self.change_type and not self.change_type in ChangeType:
            raise IllumioException("Invalid change_type: {}".format(self.change_type))
        super()._validate()


@dataclass
class ActionEvent(BaseEvent):
    task_name: str = None
    api_endpoint: str = None
    api_method: str = None
    http_status_code: int = None
    src_ip: str = None
    errors: List[Error] = None


@dataclass
@pce_api('events')
class Event(BaseEvent):
    """Represents an event object in the PCE.

    **NOTE:** Events are read-only via the PCE API.

    See https://docs.illumio.com/core/21.5/Content/Guides/rest-api/pce-management/events.htm

    Usage:
        >>> import illumio
        >>> pce = illumio.PolicyComputeEngine('pce.company.com', port=443, org_id=1)
        >>> pce.set_credentials('api_key', 'api_secret')
        >>> events = pce.events.get(params={'max_results': 5})
        >>> events
        [
            Event(
                href="/orgs/1/events/3764a636-4846-492f-a090-ae96cf33bddf",
                event_type="system_task.expire_service_account_api_keys",
                timestamp="2022-08-17T22:12:37.410Z",
                pce_fqdn="pce.company.com",
                severity="critical",
                status="success",
                created_by={"system": {}},
                action=ActionEvent(
                    uuid="1ebad14e-bd87-42b1-ae9a-22433951fbd3",
                    src_ip="FILTERED",
                    ...
                ),
                resource_changes=[],
                notifications=[],
            ),
            ...
        ]
    """
    status: str = None
    action: ActionEvent = None
    resource_changes: List[ResourceEvent] = None
    notifications: List[NotificationEvent] = None


__all__ = [
    'Event',
    'NotificationEvent',
    'ResourceEvent',
    'ActionEvent',
]



# -*- coding: utf-8 -*-

"""This module provides helper functions and decorators for common use-cases.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
import functools
import re
import socket
import sys
import typing
import warnings
from dataclasses import dataclass
from typing import Any
from urllib.parse import urlparse

from illumio._version import version
from illumio.exceptions import IllumioException, IllumioIntegerValidationException
from .constants import ACTIVE, DRAFT, PCE_APIS


def ignore_empty_keys(o: dict):
    """Removes keys with None-type values from the provided dict.

    Used for JSON encoding to avoid schema errors due to empty or invalid parameters.
    """
    return {k: v for (k, v) in o if v is not None}


def convert_draft_href_to_active(href: str) -> str:
    """Given an HREF string, converts policy version to active.

    If an active HREF is provided, this function has no effect.

    Args:
        href (str): PCE object HREF.

    Returns:
        str: active policy version HREF.
    """
    return href.replace('/{}/'.format(DRAFT), '/{}/'.format(ACTIVE))


def convert_active_href_to_draft(href: str) -> str:
    """Given an HREF string, converts policy version to draft.

    If a draft HREF is provided, this function has no effect.

    Args:
        href (str): PCE object HREF.

    Returns:
        str: draft policy version HREF.
    """
    return href.replace('/{}/'.format(ACTIVE), '/{}/'.format(DRAFT))


def deprecated(deprecated_in, message=None):
    """
    Deprecation decorator, adapted from https://stackoverflow.com/a/30253848
    Will emit a warning when the decorated function is called.
    """
    def _deprecated(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            default_message = "Call to deprecated function {}. Deprecated in version {}, current version is {}.".format(
                func.__name__, deprecated_in, version
            )
            warning_message = message or default_message
            warnings.warn(warning_message, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)
        return wrapper
    return _deprecated


def pce_api(name: str, endpoint: str = None, is_sec_policy=False, is_global=False):
    """Decorates an IllumioObject subclass to denote it as a PCE API object type.

    This registers the type in the PCE_APIS mapping used to determine whether
    a given name corresponds to an API-accessible type.

    We can then leverage __getattr__ to instantiate a generic API interface for
    any registered type (with some caveats, see the _PCEObjectAPI documentation).

    By default, registers the endpoint as /{name}, but the endpoint can also be
    specified in the decorator function call.

    For example:

    >>> @pce_api('labels', endpoint='/labels')
    >>> class Label(IllumioObject):
    ...     ...
    >>> pce = PolicyComputeEngine(...)
    >>> # the 'labels' name is registered, and so we can
    >>> # call /labels endpoints through the _PCEObjectAPI interface
    >>> labels = pce.labels.get()
    >>> labels
    [
        Label(
            href='/orgs/1/labels/1',
            key='role',
            value='R-DB',
            ...
        ),
        ...
    ]

    Args:
        name (str): the name of the API. used as a PolicyComputeEngine attribute name
            to generate the API interface.
        endpoint (str, optional): _description_. Defaults to None.
        is_sec_policy (bool, optional): whether or not the object reflects a security
            policy API with the sec_policy/{pversion} prefix. Defaults to False.
        is_global (bool, optional): whether or not the object reflects a global API,
            such as /health or /users. These APIs operate on the entire PCE rather
            than a single tenant, and don't need the /orgs/{org_id} prefix.
    """
    def _decorator(cls):
        @dataclass
        class __PCEApi:
            name: str
            endpoint: str
            object_class: object
            is_sec_policy: bool
            is_global: bool
        PCE_APIS[name] = __PCEApi(
            name=name,
            endpoint=endpoint or '/{}'.format(name),
            object_class=cls,
            is_sec_policy=is_sec_policy,
            is_global=is_global
        )
        return cls
    return _decorator


def parse_url(url: str) -> tuple:
    """Parses given URL into its scheme and hostname, stripping port and path.

    Args:
        url (str): URL to parse.

    Returns:
        tuple: parsed (scheme, hostname)
    """
    pattern = re.compile('^\w+://')
    if not re.match(pattern, url):
        url = 'https://{}'.format(url)
    parsed = urlparse(url)
    scheme = parsed.scheme
    if scheme not in ('http', 'https'):
        scheme = 'https'     # only support http(s)
    return scheme, parsed.hostname


def convert_protocol(protocol: str) -> int:
    """Given a protocol name, returns the integer ID of that protocol.

    Usage:
        >>> convert_protocol('tcp')
        6

    Args:
        protocol (str): case-insensitive protocol string, e.g. 'tcp', 'UDP'

    Raises:
        IllumioException: if an invalid protocol name is provided.

    Returns:
        int: the integer ID of the given protocol, e.g. 17 for 'udp'
    """
    try:
        return socket.getprotobyname(protocol)
    except:
        raise IllumioException('Invalid protocol name: {}'.format(protocol))


def validate_int(val: Any, minimum: int=0, maximum: int=sys.maxsize) -> None:
    """Validates a given value is an integer and is within min <= val <= max.

    Args:
        val (Any): value to validate.
        min (int, optional): validation lower bound. Defaults to 0.
        max (int, optional): validation upper bound. Defaults to sys.maxsize.

    Raises:
        IllumioIntegerValidationException: if an invalid value is provided.
    """
    try:
        valid = minimum <= int(val) <= maximum
    except:  # catch the ValueError for invalid values
        valid = False
    if not valid:
        raise IllumioIntegerValidationException(val, minimum, maximum)

if hasattr(typing, 'get_origin'):
    # python 3.8+ - introduces the get_origin function
    def isunion(type_):
        return typing.get_origin(type_) is typing.Union

    def islist(type_):
        if type_ is list:
            return True
        return typing.get_origin(type_) is list

elif hasattr(typing, '_GenericAlias'):
    # python 3.7 - changes meta types to be based off the
    # _GenericAlias supertype. __extra__ is now __origin__
    def isunion(type_):
        if isinstance(type_, typing._GenericAlias):
            return type_.__origin__ is typing.Union
        return False

    def islist(type_):
        if type_ is list:
            return True
        if isinstance(type_, typing._GenericAlias):
            return type_.__origin__ is list
        return False

else:
    # python 3.6
    def isunion(type_):
        return isinstance(type_, typing._Union)

    def islist(type_):
        if type_ is list:
            return True
        # in 3.6, List's type is GenericMeta, so we
        # instead check the __extra__ param
        if hasattr(type_, '__extra__'):
            return type_.__extra__ is list
        return False


__all__ = [
    'ignore_empty_keys',
    'convert_draft_href_to_active',
    'convert_active_href_to_draft',
    'deprecated',
    'pce_api',
    'parse_url',
    'convert_protocol',
    'validate_int',
    'isunion',
    'islist',
]



# -*- coding: utf-8 -*-

"""This module provides the PolicyComputeEngine client for communicating with the PCE.

Usage:
    >>> import illumio
    >>> pce = illumio.PolicyComputeEngine('pce.company.com', port=8443, org_id=12)
    >>> pce.set_credentials('api_key', 'api_secret')
    >>> workloads = pce.workloads.get(
    ...     params={
    ...         'managed': True,
    ...         'enforcement_mode': 'visibility_only'
    ...     }
    ... )
    >>> workloads
    [
        Workload(href='/orgs/12/workloads/c754a713-2bde-4427-af1f-bff145be509b', ...),
        ...
    ]

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
import json
import time
from typing import Any, List, Union

from requests import Session, Response
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .secpolicy import PolicyChangeset, PolicyVersion
from .exceptions import IllumioApiException
from .policyobjects import IPList, Service
from .explorer import TrafficQuery, TrafficFlow
from .util import (
    deprecated,
    convert_active_href_to_draft,
    parse_url,
    href_from,
    validate_int,
    Reference,
    IllumioEncoder,
    ACTIVE,
    DRAFT,
    PORT_MAX,
    ANY_IP_LIST_NAME,
    ALL_SERVICES_NAME,
    BULK_CHANGE_LIMIT,
    PCE_APIS
)


class PolicyComputeEngine:
    """The REST client core for the Illumio Policy Compute Engine.

    Contains request logic for API calls and handles the HTTP(S) connection to the PCE.

    Usage:
        >>> import illumio
        >>> pce = illumio.PolicyComputeEngine('pce.company.com', port=8443, org_id=12)
        >>> pce.set_credentials('api_key', 'api_secret')
        >>> workloads = pce.workloads.get(
        ...     params={
        ...         'managed': True,
        ...         'enforcement_mode': 'visibility_only'
        ...     }
        ... )
        >>> workloads
        [
            Workload(href='/orgs/12/workloads/c754a713-2bde-4427-af1f-bff145be509b', ...),
            ...
        ]

    Args:
        url (str): PCE URL. May include http:// or https:// as the scheme.
        port (str, optional): PCE http(s) port. Defaults to '443'.
        version (str, optional): The PCE API version to use. Defaults to 'v2'.
        org_id (str, optional): The PCE organization ID. Defaults to '1'.
        retry_count (int, optional): Number of times to retry on failure.
            Defaults to 5.
        request_timeout (int, optional): HTTP request timeout in seconds.
            Defaults to 30.

    Attributes:
        base_url: DEPRECATED in v1.0.3. The base URL for API calls to the PCE.
            Has the form ``http[s]://<DOMAIN_NAME>:<PORT>/api/<API_VERSION>``
        include_org: flag denoting whether to prepend the orgs subpath
            to request endpoints by default. Defaults to True.
        org_id: the PCE organization ID.
    """
    def __init__(self, url: str, port: str = '443', version: str = 'v2', org_id: str = '1',
                    retry_count: int = 5, request_timeout: int = 30) -> None:
        self._apis = {}
        self._encoder = IllumioEncoder()
        self._session = Session()
        self._session.headers.update({'Accept': 'application/json'})
        self._scheme, self._hostname = parse_url(url)
        self._port = port
        self._version = version
        self._timeout = request_timeout
        # leaving this in for backwards compatibility
        self.base_url = "{}://{}:{}/api/{}".format(
            self._scheme, self._hostname, port, version
        )
        self.include_org = True
        self.org_id = org_id
        self._validate()
        self._setup_retry(retry_count)

    def _validate(self) -> None:
        """Validates configuration values, raising an error on failure"""
        validate_int(self._port, minimum=1, maximum=PORT_MAX)
        validate_int(self.org_id, minimum=1)

    def _setup_retry(self, retries: int) -> None:
        """Configures `requests.Session` retry defaults"""
        retry_strategy = Retry(
            total=retries,
            # {backoff} * (2 ** ({retry count} - 1))
            # 1, 2, 4, 8, 16 (seconds)
            backoff_factor=2,
            status_forcelist=[429, 500, 502, 503, 504]
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self._session.mount("https://", adapter)
        self._session.mount("http://", adapter)

    def set_credentials(self, username: str, password: str) -> None:
        """Sets username and password values to authenticate with the PCE.

        The provided credentials can be either user credentials or an API
        key/secret pair.

        Args:
            username (str): username or API key.
            password (str): password or API secret.
        """
        self._session.auth = (username, password)

    def set_proxies(self, http_proxy: str = None, https_proxy: str = None) -> None:
        """Sets HTTP proxies to use when connecting to the PCE.

        Args:
            http_proxy (str, optional): HTTP proxy URI. Defaults to None.
            https_proxy (str, optional): HTTPS proxy URI. Defaults to None.
        """
        self._session.proxies.update({'http': http_proxy, 'https': https_proxy})

    def set_timeout(self, timeout: int) -> None:
        """Sets the HTTP request timeout for PCE connections.

        Args:
            timeout (int): the new request timeout in seconds.
        """
        self._timeout = timeout

    def _request(self, method: str, endpoint: str, include_org: bool = None, **kwargs) -> Response:
        """Makes an API call to the PCE.

        Args:
            method (str): the HTTP request method. Supports the same verbs as
                `requests.request`.
            endpoint (str): the API endpoint to call.
            include_org (bool, optional): whether or not to include /orgs/{org_id}
                in the API call. Defaults to the value of the `include_org` class
                attribute.

        Raises:
            IllumioApiException: if the response is unsuccessful (status code >399).

        Returns:
            requests.Response: the `Response` object returned from a successful request.
        """
        try:
            include_org = self.include_org if include_org is None else include_org
            # avoid reference before assignment errors in case of cxn failure
            response = None
            url = self._build_url(endpoint, include_org)
            self._encode_body(kwargs)
            kwargs['timeout'] = kwargs.get('timeout', self._timeout)
            response = self._session.request(method, url, **kwargs)
            response.raise_for_status()
            return response
        except Exception as e:
            message = str(e)
            content_type = response.headers.get('Content-Type', '')
            # Response objects are falsy if the request failed so do a null check
            if response is not None and content_type == 'application/json':
                message = self._get_error_message_from_response(response)
            raise IllumioApiException(message) from e

    def _build_url(self, endpoint: str, include_org: bool):
        endpoint = endpoint.lstrip('/').replace('//', '/')
        if include_org and not endpoint.startswith('orgs/'):
            endpoint = 'orgs/{}/{}'.format(self.org_id, endpoint)
        return '{}://{}:{}/api/{}/{}'.format(
            self._scheme, self._hostname, self._port, self._version, endpoint
        )

    def _encode_body(self, kwargs):
        """Encodes request body data to JSON."""
        body = kwargs.pop('data', None)
        if 'json' in kwargs:
            # json overrides data if both are provided
            body = kwargs.pop('json')
        if body is not None:
            kwargs['json'] = json.loads(self._encoder.encode(body))

    def _get_error_message_from_response(self, response: Response) -> str:
        message = "API call returned error code {}. Errors:".format(response.status_code)
        for error in response.json():
            if error and 'token' in error and 'message' in error:
                message += '\n{}: {}'.format(error['token'], error['message'])
            elif error and 'error' in error:
                message += '\n{}'.format(error['error'])
        return message

    def get(self, endpoint: str, **kwargs) -> Response:
        """Makes a GET call to a given PCE endpoint.

        Additional keyword arguments are passed to the `requests.Request` object.

        Args:
            endpoint (str): the PCE endpoint to call.

        Returns:
            requests.Response: the `Response` object returned from a successful request.
        """
        return self._request('GET', endpoint, **kwargs)

    def post(self, endpoint: str, **kwargs) -> Response:
        """Makes a POST call to a given PCE endpoint.

        Appends 'Content-Type: application/json' to the request headers by default.
        Additional keyword arguments are passed to the `requests.Request` object.

        Args:
            endpoint (str): the PCE endpoint to call.

        Returns:
            requests.Response: the `Response` object returned from a successful request.
        """
        headers = kwargs.get('headers', {})
        kwargs['headers'] = {**headers, **{'Content-Type': 'application/json'}}
        return self._request('POST', endpoint, **kwargs)

    def put(self, endpoint: str, **kwargs) -> Response:
        """Makes a PUT call to a given PCE endpoint.

        Appends 'Content-Type: application/json' to the request headers by default.
        Additional keyword arguments are passed to the `requests.Request` object.

        Args:
            endpoint (str): the PCE endpoint to call.

        Returns:
            requests.Response: the `Response` object returned from a successful request.
        """
        headers = kwargs.get('headers', {})
        kwargs['headers'] = {**headers, **{'Content-Type': 'application/json'}}
        return self._request('PUT', endpoint, **kwargs)

    def delete(self, endpoint: str, **kwargs) -> Response:
        """Makes a DELETE call to a given PCE endpoint.

        Additional keyword arguments are passed to the `requests.Request` object.

        Args:
            endpoint (str): the PCE endpoint to call.

        Returns:
            requests.Response: the `Response` object returned from a successful request.
        """
        return self._request('DELETE', endpoint, **kwargs)

    def get_collection(self, endpoint: str, **kwargs) -> Response:
        """Uses the PCE's asynchronous job logic to retrieve a collection of objects.

        **NOTE:** for large collections (surpassing 10,000 objects), this function
        will be extremely slow - it is recommended that callers use query filters
        or the max_results parameter to limit the number of results in the collection.

        Args:
            endpoint (str): the PCE endpoint to call.

        Raises:
            IllumioApiException: if an error message (>399) status code is returned.

        Returns:
            requests.Response: the `Response` object returned from a successful request.
        """
        try:
            headers = kwargs.get('headers', {})
            kwargs['headers'] = {**headers, **{'Prefer': 'respond-async'}}
            response = self.get(endpoint, **kwargs)
            response.raise_for_status()
            location = response.headers['Location']
            retry_after = int(response.headers['Retry-After'])

            collection_href = self._async_poll(location, retry_after)

            response = self.get(collection_href)
            response.raise_for_status()
            return response
        except Exception as e:
            raise IllumioApiException from e

    def _async_poll(self, job_location: str, retry_time: Union[int, float] = 1.) -> str:
        """Polls the PCE for an async job's status until it completes or times out.

        The poll-wait loop uses the Retry-After time from the job submission request
        or a default of 1 second, with a 1.5x backoff.

        Args:
            job_location (str): URL of the job to poll.
            retry_time (Union[int, float], optional): base Retry-After time.
                Defaults to 1 second.

        Raises:
            Exception: if the job returns a 'failed' status.

        Returns:
            str: the HREF path of the completed collection document.
        """
        while True:
            time.sleep(retry_time)
            retry_time *= 1.5  # slight backoff to avoid spamming the PCE for long-running jobs
            response = self.get(job_location)
            response.raise_for_status()
            poll_result = response.json()
            poll_status = poll_result['status']

            if poll_status == 'failed':
                raise Exception('Async collection job failed: ' + poll_result['result']['message'])
            elif poll_status == 'completed':
                # traffic flow async jobs
                collection_href = poll_result['result']
                break
            elif poll_status == 'done':
                # policy object collection jobs
                collection_href = poll_result['result']['href']
                break
        return collection_href

    def check_connection(self, **kwargs) -> bool:
        """Makes a GET request to the PCE /health endpoint.

        Additional keyword arguments are passed to the requests call.

        Returns:
            bool: True if the call is successful, otherwise False.
        """
        try:
            self.get('/health', **{**kwargs, **{'include_org': False}})
            # make an /orgs/{org_id} call to validate the org ID as well
            # /settings/workloads is a relatively quick call that will work on SaaS PCEs
            self.get('/settings/workloads', **{**kwargs, **{'include_org': True}})
            return True
        except IllumioApiException:
            return False

    class _PCEObjectAPI:
        """Generic API for registered PCE objects.

        Each registered API exposes CRUD operation functions through this common interface.
        """
        def __init__(self, pce: 'PolicyComputeEngine', api_data: object) -> None:
            self.name = api_data.name
            self.endpoint = api_data.endpoint
            self.object_cls = api_data.object_class
            self.is_sec_policy = api_data.is_sec_policy
            self.is_global = api_data.is_global
            self.pce = pce

        def _build_endpoint(self, policy_version: str, parent: Any) -> str:
            """Builds the PCE request endpoint."""
            endpoint = self.endpoint

            if parent:  # e.g. /sec_policy/active/rulesets/1/sec_rules
                parent_draft_href = convert_active_href_to_draft(href_from(parent))
                endpoint = '{}/{}'.format(parent_draft_href, endpoint)
            else:  # mutually exclusive as the parent HREF will have the sec_policy and orgs prefix already
                if self.is_sec_policy:
                    if policy_version not in [ACTIVE, DRAFT]:
                        raise IllumioApiException("Invalid policy_version passed to get: {}".format(policy_version))
                    endpoint = '/sec_policy/{}/{}'.format(policy_version, endpoint)

                if not self.is_global:
                    endpoint = '/orgs/{}/{}'.format(self.pce.org_id, endpoint)
            return endpoint.replace('//', '/')

        def get_by_reference(self, reference: Union[str, Reference, dict], **kwargs) -> Reference:
            """Retrieves an object from the PCE using its HREF.

            Usage:
                >>> ip_list = pce.ip_lists.get_by_reference('/orgs/1/sec_policy/active/ip_lists/1')
                >>> ip_list
                IPList(
                    name='Any (0.0.0.0/0 and ::/0)',
                    ...
                )

            Args:
                href (str): the HREF of the object to fetch.

            Returns:
                Reference: the object json, decoded to its python equivalent.
            """
            kwargs['include_org'] = False
            response = self.pce.get(href_from(reference), **kwargs)
            return self.object_cls.from_json(response.json())

        def get_by_name(self, name: str, policy_version: str = DRAFT, **kwargs) -> Reference:
            """Retrieves the object from the PCE with the given name.

            Args:
                name (str): the exact name of the object to retrieve.

            Returns:
                Reference: the decoded object, or None if an object with the
                    given name wasn't found.
            """
            kwargs['params'] = {'name': name}
            endpoint = self._build_endpoint(policy_version, None)
            response = self.pce.get(endpoint, **kwargs)
            for o in response.json():
                if 'name' in o and o['name'] == name:
                    return self.object_cls.from_json(o)

        def get(self, policy_version: str = DRAFT, parent: Union[str, Reference, dict] = None, **kwargs) -> List[Reference]:
            """Retrieves objects from the PCE based on the given parameters.

            Keyword arguments to this function are passed to the `requests.get` call.
            See https://docs.illumio.com/core/21.5/API-Reference/index.html
            for details on filter parameters for collection queries.

            Usage:
                >>> virtual_services = pce.virtual_services.get(
                ...     policy_version='active',
                ...     params={
                ...         'name': 'VS-'
                ...     }
                ... )
                >>> virtual_services
                [
                    VirtualService(
                        href='/orgs/1/sec_policy/active/virtual_services/9177c75f-7b21-4bf0-8c16-2c47c1ca3252',
                        name='VS-LAB-SERVICES'
                        ...
                    ),
                    ...
                ]

            Args:
                policy_version (str, optional): if fetching security policy objects, specifies
                    whether to fetch 'draft' or 'active' objects. Defaults to 'draft'.
                parent (Union[str, Reference, dict], optional): Reference to the
                    object's parent. Required for some object types, such
                    as Security Rules which must be created as children of
                    existing RuleSets.

            Returns:
                List[Reference]: the returned list of decoded objects.
            """
            endpoint = self._build_endpoint(policy_version, parent)
            response = self.pce.get(endpoint, **{**kwargs, **{'include_org': False}})
            return [self.object_cls.from_json(o) for o in response.json()]

        def get_all(self, policy_version: str = DRAFT, parent: Union[str, Reference, dict] = None, **kwargs) -> List[Reference]:
            """Retrieves all objects of a given type from the PCE.

            This function makes two requests, using the `X-Total-Count` header
            in the response to set the `max_results` parameter on the follow-up
            request.

            Args:
                policy_version (str, optional): if fetching security policy objects, specifies
                    whether to fetch 'draft' or 'active' objects. Defaults to 'draft'.
                parent (Union[str, Reference, dict], optional): Reference to the
                    object's parent. Required for some object types, such
                    as Security Rules which must be created as children of
                    existing RuleSets.

            Returns:
                List[Reference]: the returned list of decoded objects.
            """
            kwargs['include_org'] = False
            params = kwargs.get('params', {})
            endpoint = self._build_endpoint(policy_version, parent)

            if 'max_results' not in params:
                kwargs['params'] = {**params, **{'max_results': 0}}
                response = self.pce.get(endpoint, **kwargs)
                if len(response.json()) > 0:  # for endpoints that don't support max_results
                    return [self.object_cls.from_json(o) for o in response.json()]
                filtered_object_count = response.headers['X-Total-Count']
                kwargs['params'] = {**params, **{'max_results': int(filtered_object_count)}}

            response = self.pce.get(endpoint, **kwargs)
            return [self.object_cls.from_json(o) for o in response.json()]

        def get_async(self, policy_version: str = DRAFT, parent: Union[str, Reference, dict] = None, **kwargs) -> List[Reference]:
            """Retrieves objects asynchronously from the PCE based on the given parameters.

            Args:
                policy_version (str, optional): if fetching security policy objects, specifies
                    whether to fetch 'draft' or 'active' objects. Defaults to 'draft'.
                parent (Union[str, Reference, dict], optional): Reference to the
                    object's parent. Required for some object types, such
                    as Security Rules which must be created as children of
                    existing Rule Sets.

            Returns:
                List[Reference]: the returned list of decoded objects.
            """
            kwargs['include_org'] = False
            endpoint = self._build_endpoint(policy_version, parent)
            response = self.pce.get_collection(endpoint, **kwargs)
            return [self.object_cls.from_json(o) for o in response.json()]

        def create(self, body: Any, parent: Union[str, Reference, dict] = None, **kwargs) -> Reference:
            """Creates an object in the PCE.

            See https://docs.illumio.com/core/21.5/API-Reference/index.html
            for details on POST body parameters when creating objects.

            Usage:
                >>> from illumio.policyobjects import Label
                >>> label = Label(key='role', value='R-DB')
                >>> label = pce.labels.create(label)
                >>> label
                Label(
                    'href': '/orgs/1/labels/14',
                    'key': 'role',
                    'value': 'R-DB
                )

            Args:
                body (Any): the parameters for the newly created object.
                parent (Union[str, Reference, dict], optional): Reference to the
                    object's parent. Required for some object types, such
                    as Security Rules which must be created as children of
                    existing RuleSets.

            Returns:
                Reference: the created object.
            """
            kwargs = {**kwargs, **{'json': body, 'include_org': False}}
            endpoint = self._build_endpoint(DRAFT, parent)
            response = self.pce.post(endpoint, **kwargs)
            return self._parse_response_body(response.json())

        def _parse_response_body(self, json_response):
            # XXX: workaround for Service Bindings. Multiple bindings
            #   can be created in the same POST, so we need to accommodate
            #   this case by checking the response body type
            if type(json_response) is list:
                results = {self.name: [], 'errors': []}
                for o in json_response:
                    if 'href' in o:
                        results[self.name].append(self.object_cls.from_json(o))
                    else:
                        results['errors'].append(o)
                return results
            return self.object_cls.from_json(json_response)

        def update(self, reference: Union[str, Reference, dict], body: Any, **kwargs) -> None:
            """Updates an object in the PCE.

            Successful PUT requests return a 204 No Content response.

            Usage:
                >>> pairing_profiles = pce.pairing_profile.get_by_name('PP-Database')
                >>> existing_profile = pairing_profiles[0]
                >>> update = PairingProfile(
                ...     name='PP-DATABASE-VENS',
                ...     enabled=False  # disable this profile
                ... )
                >>> pce.pairing_profile.update(existing_profile['href'], update)

            Args:
                reference (Union[str, Reference, dict]): the HREF of the pairing
                    profile to update.
                body (Any): the update data.
            """
            kwargs['json'] = body
            kwargs['include_org'] = False
            self.pce.put(href_from(reference), **kwargs)

        def delete(self, reference: Union[str, Reference, dict], **kwargs) -> None:
            """Deletes an object in the PCE.

            Successful DELETE requests return a 204 No Content response.

            Args:
                reference (Union[str, Reference, dict]): the HREF of the object to delete.
            """
            self.pce.delete(href_from(reference), **{**kwargs, **{'include_org': False}})

        def _bulk_change(self, objects: List[Reference], method: str, success_status: str, **kwargs) -> List[dict]:
            results = []
            kwargs['include_org'] = False
            while objects:
                kwargs['json'] = objects[:BULK_CHANGE_LIMIT]
                objects = objects[BULK_CHANGE_LIMIT:]
                endpoint = self._build_endpoint(DRAFT, None)
                response = self.pce.put('{}/{}'.format(endpoint, method), **kwargs)
                for result in response.json():
                    errors = result.get('errors', [])
                    if success_status and result['status'] != success_status:
                        errors.append({'token': result['token'], 'message': result['message']})
                    results.append({'href': result['href'], 'errors': errors})
            return results

        def bulk_create(self, objects_to_create: List[Reference], **kwargs) -> List[dict]:
            """Creates a set of objects in the PCE.

            **NOTE:** Bulk creation can currently only be applied for Security Principals,
                Virtual Services and Workloads.

            Args:
                objects_to_create (List[Reference]): list of objects to update.

            Returns:
                List[dict]: a list containing HREFs of created objects
                    as well as any errors returned from the PCE.
                    Has the following form:

                    >>> [
                    ...     {
                    ...         'href': {object_href},
                    ...         'errors': [
                    ...              {
                    ...                 'token': {error_type},
                    ...                 'message': {error_message}
                    ...             }
                    ...         ]
                    ...     }
                    ... ]
            """
            return self._bulk_change(objects_to_create, method='bulk_create', success_status='created', **kwargs)

        def bulk_update(self, objects_to_update: List[Reference], **kwargs) -> List[dict]:
            """Updates a set of objects in the PCE.

            **NOTE:** Bulk updates can currently only be applied for Virtual Services and Workloads.

            Args:
                objects_to_update (List[Reference]): list of objects to update.

            Returns:
                List[dict]: a list containing HREFs of updated objects
                    as well as any errors returned from the PCE.
                    Has the following form:

                    >>> [
                    ...     {
                    ...         'href': {object_href},
                    ...         'errors': [
                    ...             {
                    ...                 'token': {error_type},
                    ...                 'message': {error_message}
                    ...             }
                    ...         ]
                    ...     }
                    ... ]
            """
            return self._bulk_change(objects_to_update, method='bulk_update', success_status='updated', **kwargs)

        def bulk_delete(self, refs: List[Union[str, Reference, dict]], **kwargs) -> List[dict]:
            """Deletes a set of objects in the PCE.

            **NOTE:** Bulk updates can currently only be applied for Workloads.

            Args:
                hrefs (List[Union[str, Reference, dict]]): list of references to objects to delete.

            Returns:
                List[dict]: a list containing any errors that occurred during
                    the bulk operation. Has the following form:

                    >>> [
                    ...     {
                    ...         'href': {object_href},
                    ...         'errors': [
                    ...             {
                    ...                 'token': {error_type},
                    ...                 'message': {error_message}
                    ...             }
                    ...         ]
                    ...     }
                    ... ]
            """
            objects_to_delete = [Reference(href=href_from(reference)) for reference in refs]
            return self._bulk_change(objects_to_delete, method='bulk_delete', success_status=None, **kwargs)

    def __getattr__(self, name: str) -> _PCEObjectAPI:
        """Instantiates a generic API for registered PCE objects.

        Inspired by the Zabbix API: https://pypi.org/project/zabbix-api/
        """
        if name in self._apis:
            return self._apis[name]
        if name not in PCE_APIS:
            raise AttributeError("No such PCE API object: {}".format(name))
        api = self._PCEObjectAPI(pce=self, api_data=PCE_APIS[name])
        self._apis[name] = api
        return api

    def get_default_ip_list(self, **kwargs) -> IPList:
        """Retrieves the "Any (0.0.0.0/0 and ::/0)" default global IP list.

        Returns:
            IPList: decoded object representing the default global IP list.
        """
        params = kwargs.get('params', {})
        # retrieve by name as each org will use a different ID
        kwargs['params'] = {**params, **{'name': ANY_IP_LIST_NAME}}
        kwargs['include_org'] = True
        response = self.get('/sec_policy/active/ip_lists', **kwargs)
        return IPList.from_json(response.json()[0])

    def get_default_service(self, **kwargs) -> Service:
        """Retrieves the "All Services" default global Service.

        Returns:
            Service: decoded object representing the default global Service.
        """
        params = kwargs.get('params', {})
        # retrieve by name as each org will use a different ID
        kwargs['params'] = {**params, **{'name': ALL_SERVICES_NAME}}
        kwargs['include_org'] = True
        response = self.get('/sec_policy/active/services', **kwargs)
        return Service.from_json(response.json()[0])

    def generate_pairing_key(self, pairing_profile_href: str, **kwargs) -> str:
        """Generates a pairing key using a pairing profile.

        Args:
            pairing_profile_href (str): the HREF of the pairing profile to use.

        Returns:
            str: the pairing key value.
        """
        response = self.post('{}/pairing_key'.format(pairing_profile_href), **{**kwargs, **{'json': {}}})
        return response.json().get('activation_code')

    @deprecated(deprecated_in='1.0.0')
    def get_traffic_flows(self, traffic_query: TrafficQuery, **kwargs) -> List[TrafficFlow]:
        """DEPRECATED (v1.0.0). Use `get_traffic_flows_async` instead.

        Retrieves Explorer traffic flows using the provided query.

        **NOTE:** this function is deprecated in the Illumio REST API, and is
        only provided for compatibility. The Illumio Explorer REST API
        documentation recommends using the async traffic flow query instead,
        provided here as `PolicyComputeEngine.get_traffic_flows_async`.

        See https://docs.illumio.com/core/21.5/Content/Guides/rest-api/visualization/explorer.htm#TrafficAnalysisQueries
        for details on traffic query parameters.

        Args:
            traffic_query (TrafficQuery): `TrafficQuery` object representing
                the query parameters.

        Returns:
            List[TrafficFlow]: list of `TrafficFlow` objects found using the
                provided query.
        """
        kwargs = {**kwargs, **{'json': traffic_query, 'include_org': True}}
        response = self.post('/traffic_flows/traffic_analysis_queries', **kwargs)
        return [TrafficFlow.from_json(flow) for flow in response.json()]

    def get_traffic_flows_async(self, query_name: str, traffic_query: TrafficQuery, **kwargs) -> List[TrafficFlow]:
        """Retrieves Explorer traffic flows using the provided query.

        See https://docs.illumio.com/core/21.5/Content/Guides/rest-api/visualization/explorer.htm#AsynchronousQueriesforTrafficFlows
        for details on async traffic query parameters.

        Usage:
            >>> traffic_query = TrafficQuery.build(
            ...     start_date="2022-02-01T00:00:00Z",
            ...     end_date="2022-03-01T00:00:00Z",
            ...     include_services=[
            ...         {'port': 3389, 'proto': 'tcp'}
            ...     ],
            ...     policy_decisions=['potentially_blocked', 'unknown']
            ... )
            >>> traffic_query
            TrafficQuery(
                start_date='2022-02-01T00:00:00Z',
                end_date='2022-03-01T00:00:00Z',
                sources=TrafficQueryFilterBlock(
                    include=[],
                    exclude=[]
                ),
                destinations=TrafficQueryFilterBlock(
                    include=[],
                    exclude=[]
                ),
                services=TrafficQueryServiceBlock(
                    include=[
                        ServicePort(
                            port=3389,
                            proto=6,
                            ...
                        )
                    ],
                    exclude=[]
                ),
                policy_decisions=[
                    'potentially_blocked',
                    'unknown'
                ],
                ...
            )
            >>> traffic_flows = pce.get_traffic_flows_async(
            ...     query_name='rdp-traffic-feb-22',
            ...     traffic_query=traffic_query
            ... )
            >>> traffic_flows
            [
                TrafficFlow(
                    src=TrafficNode(
                        ip='10.0.9.14',
                        "ip_lists": [
                            {
                                "name": "IPL-INTERNAL",
                                "href": "/orgs/1/sec_policy/active/ip_lists/5",
                                ...
                            }
                        ],
                        ...
                    ),
                    dst=TrafficNode(
                        ip='10.0.6.63',
                        workload=Workload(
                            href='/orgs/1/workloads/6567900a-b49f-43cc-93a7-c892da39aad1',
                            name='WIN-JUMPBOX',
                            ...
                        ),
                        ...
                    ),
                    service=ServicePort(port=3389, proto=6),
                    num_connections=44,
                    state='closed',
                    timestamp_range=TimestampRange(
                        first_detected='2022-02-19T09:50:17Z',
                        last_detected='2022-02-19T10:12:36Z'
                    ),
                    policy_decision='potentially_blocked',
                    flow_direction='inbound',
                    ...
                ),
                ...
            ]

        Args:
            query_name (str): name for the async query job.

            traffic_query (TrafficQuery): `TrafficQuery` object representing
                the query parameters.

        Raises:
            IllumioApiException: if there is an error retrieving the async job
                results.

        Returns:
            List[TrafficFlow]: list of `TrafficFlow` objects found using the
                provided query.
        """
        try:
            traffic_query.query_name = query_name
            kwargs['json'] = traffic_query
            headers = kwargs.get('headers', {})
            kwargs['headers'] = {**headers, **{
                'Content-Type': 'application/json',
                'Prefer': 'respond-async'
            }}
            kwargs['include_org'] = True
            response = self.post('/traffic_flows/async_queries', **kwargs)
            response.raise_for_status()
            query_status = response.json()
            location = query_status['href']

            collection_href = self._async_poll(location)

            response = self.get(collection_href)
            response.raise_for_status()
            return [TrafficFlow.from_json(flow) for flow in response.json()]
        except Exception as e:
            raise IllumioApiException from e

    def provision_policy_changes(self, change_description: str, hrefs: List[str], **kwargs) -> PolicyVersion:
        """Provisions policy changes for draft objects with the given HREFs.

        Usage:
            >>> rule_set = pce.rule_sets.create(
            ...     RuleSet(name='RS-RINGFENCE')
            ... )
            >>> changeset = pce.provision_policy_changes(
            ...     change_description='Provision ring-fence rule set',
            ...     hrefs=[rule_set.href]
            ... )
            >>> changeset
            PolicyVersion(
                href='/orgs/1/sec_policy/110',
                commit_message='Provision ring-fence rule set',
                version=110,
                workloads_affected=0,
                object_counts=PolicyObjectCounts(
                    label_groups=17,
                    services=8,
                    ...
                )
            )

        Args:
            change_description (str): the policy change description.
            hrefs (List[str]): the HREFs of the draft policy objects to provision.

        Raises:
            IllumioException: if an invalid HREF is provided.

        Returns:
            PolicyVersion: the decoded policy version object including the changeset.
        """
        policy_changeset = PolicyChangeset.build(hrefs)
        kwargs['json'] = {
            'update_description': change_description,
            'change_subset': policy_changeset
        }
        response = self.post('/sec_policy', **{**kwargs, **{'include_org': True}})
        return PolicyVersion.from_json(response.json())


__all__ = ['PolicyComputeEngine']



# -*- coding: utf-8 -*-

"""This module is a stub for container cluster objects.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass
from typing import List

from illumio.exceptions import IllumioException
from illumio.util import (
    JsonObject,
    Reference,
    IllumioObject,
    MutableObject,
    EnforcementMode,
    VisibilityLevel,
    pce_api
)


@dataclass
class ContainerClusterNode(JsonObject):
    name: str = None
    pod_subnet: str = None


@dataclass
class ContainerClusterError(JsonObject):
    audit_event: Reference = None
    duplicate_ids: List[str] = None
    error_type: str = None


@dataclass
class LabelRestriction(JsonObject):
    """Represents a label or set of labels to restrict the workload profile to.

    Only one of ``assignment`` or ``restriction`` can be specified.

    Args:
        key (str, optional): label type, for example 'role', 'app'.
        assignment (Reference, optional): reference to label object to assign
            to container workloads using the profile.
        restriction (List[Reference], optional): restricts container workload
            label assignment for the given key to only the referenced labels.
    """
    key: str = None
    assignment: Reference = None
    restriction: List[Reference] = None


@dataclass
@pce_api('container_workload_profiles')
class ContainerWorkloadProfile(MutableObject):
    """Represents a workload profile within a container cluster object in the PCE.

    Workload profiles define management and scope for container workloads under
    a cluster namespace defined by the profile.

    The `assign_labels` field is DEPRECATED in favour of the more flexible `labels`
    for defining label assignments and restrictions on the profile. `assign_labels`
    is left in for compatibility with older PCE versions. In either case, label
    assignments can only be specified for managed workload profiles.

    **NOTE:** though the `enforcement_mode` value for a workload profile can be
    set to `selective`, it is currently not supported and may result in
    unexpected behaviour. The only supported enforcement modes for workload
    profiles are ``idle``, ``visibility_only``, and ``full``.

    Usage:
        >>> import illumio
        >>> pce = illumio.PolicyComputeEngine('pce.company.com', port=443, org_id=1)
        >>> pce.set_credentials('api_key', 'api_secret')
        >>> container_cluster = illumio.ContainerCluster(
        ...     name='CC-EKS-PROD',
        ...     description='Production Kubernetes cluster on AWS'
        ... )
        >>> container_cluster = pce.container_clusters.create(container_cluster)
        >>> env_label = pce.labels.create({'key': 'env', 'value': 'Production'})
        >>> loc_label = pce.labels.create({'key': 'loc', 'value': 'AWS'})
        >>> container_workload_profile = illumio.ContainerWorkloadProfile(
        ...     name='illumio-system',
        ...     managed=True,
        ...     labels=[
        ...         illumio.LabelRestriction(key='env', assignment=env_label),
        ...         illumio.LabelRestriction(key='loc', assignment=loc_label)
        ...     ],
        ...     enforcement_mode='visibility_only'
        ... )
        >>> container_workload_profile = pce.container_workload_profiles.create(
        ...     container_workload_profile, parent=container_cluster
        ... )
        >>> container_workload_profile
        ContainerWorkloadProfile(
            href='/orgs/1/container_clusters/f5bef182-8c55-4219-b35b-0a50b707e434/container_workload_profiles/d2d466b5-106d-48e9-ada9-68f6321d1da8',
            name='illumio-system',
            namespace=None,
            managed=True,
            labels=[
                LabelRestriction(
                    key='env',
                    assignment=Reference(href='/orgs/1/labels/23'),
                    ...
                ),
                ...
            ],
            enforcement_mode='visibility_only'
            ...
        )
    """
    namespace: str = None
    assign_labels: List[Reference] = None
    labels: List[LabelRestriction] = None
    enforcement_mode: str = None
    visibility_level: str = None
    linked: bool = None
    managed: bool = None

    def _validate(self):
        if self.enforcement_mode and self.enforcement_mode not in EnforcementMode:
            raise IllumioException("Invalid enforcement_mode: {}".format(self.enforcement_mode))
        if self.visibility_level and self.visibility_level not in VisibilityLevel:
            raise IllumioException("Invalid visibility_level: {}".format(self.visibility_level))
        return super()._validate()


@dataclass
@pce_api('container_clusters')
class ContainerCluster(IllumioObject):
    """Represents a container cluster object in the PCE.

    Container clusters are abstract representations of container orchestration
    systems linked to the PCE.

    **NOTE:** when a container cluster is created through the API, the
    `container_cluster_token` used by Kubelink and C-VEN containers
    to pair with the PCE is returned in the response. This token is
    only available after the initial POST request and cannot be
    retrieved via the API: make sure to store it in a persistent
    form after creating the cluster.

    See https://docs.illumio.com/core/21.5/Content/LandingPages/Guides/kubernetes-and-openshift.htm

    Usage:
        >>> import illumio
        >>> pce = illumio.PolicyComputeEngine('pce.company.com', port=443, org_id=1)
        >>> pce.set_credentials('api_key', 'api_secret')
        >>> container_cluster = illumio.ContainerCluster(
        ...     name='CC-EKS-PROD',
        ...     description='Production Kubernetes cluster on AWS'
        ... )
        >>> container_cluster = pce.container_clusters.create(container_cluster)
        >>> container_cluster
        ContainerCluster(
            href='/orgs/1/container_clusters/f5bef182-8c55-4219-b35b-0a50b707e434',
            name='CC-EKS-PROD',
            description='Production Kubernetes cluster on AWS',
            container_cluster_token='1_016dace1ab35fafe8e71c6dda6695e0881393f1f4c494e6cd70178f1e743b372',
            ...
        )
    """
    pce_fqdn: str = None
    manager_type: str = None
    last_connected: str = None
    kubelink_version: str = None
    online: bool = None
    nodes: List[ContainerClusterNode] = None
    container_runtime: str = None
    errors: List[ContainerClusterError] = None
    container_cluster_token: str = None


__all__ = [
    'ContainerClusterNode',
    'ContainerClusterError',
    'LabelRestriction',
    'ContainerWorkloadProfile',
    'ContainerCluster',
]



# -*- coding: utf-8 -*-

"""This module provides constant values and enumerations used by the PCE REST API.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
import re
from enum import Enum, EnumMeta

#: Active policy version path literal.
ACTIVE = 'active'

#: Draft policy version path literal.
DRAFT = 'draft'

#: Used in rules and enforcement boundaries to denote that all
#: workloads should be affected.
AMS = 'ams'

#: Used in resolve_labels_as block in rule creation to denote that
#: workloads matching the rule scope should be affected.
RESOLVE_AS_WORKLOADS = 'workloads'

#: Used in resolve_labels_as block in rule creation to denote that
#: virtual services matching the rule scope should be affected.
RESOLVE_AS_VIRTUAL_SERVICES = 'virtual_services'

#: Name of the default global IP list.
ANY_IP_LIST_NAME = 'Any (0.0.0.0/0 and ::/0)'

#: Name of the default global Service.
ALL_SERVICES_NAME = 'All Services'

#: Max port number.
PORT_MAX = 65535

#: Max value for the ICMP header Code field.
#: See https://www.iana.org/assignments/icmp-parameters/icmp-parameters.xhtml#icmp-parameters-codes
ICMP_CODE_MAX = 15

#: Max value for the ICMP header Type field.
#: See https://www.iana.org/assignments/icmp-parameters/icmp-parameters.xhtml#icmp-parameters-types
ICMP_TYPE_MAX = 255

FQDN_REGEX = re.compile('(?=^.{4,253}$)(^((?!-)[a-zA-Z0-9-]{1,63}(?<!-)\.)+[a-zA-Z]{2,63}$)')
HREF_REGEX = re.compile('^\/orgs\/\d+\/(?:sec_policy\/(?:active|draft)\/)?(?P<type>[a-zA-Z_]+)\/(?P<uid>[a-zA-Z0-9-]+)$')

#: Upper limit on the number of objects that can be sent to PCE bulk change
#: endpoints in a single request.
BULK_CHANGE_LIMIT = 1000

PCE_APIS = {}


class IllumioEnumMeta(EnumMeta):
    """Metaclass to provide a common contains check for enumerations."""

    def __contains__(cls, value):
        if value is None:
            return False
        if type(value) is str:
            value = value.lower()
        if isinstance(type(value), IllumioEnumMeta):
            value = value.value
        return value in cls._value2member_map_


class LinkState(str, Enum, metaclass=IllumioEnumMeta):
    """Network interface link state enumeration."""
    UP = 'up'
    DOWN = 'down'
    UNKNOWN = 'unknown'


class EnforcementMode(str, Enum, metaclass=IllumioEnumMeta):
    """Workload enforcement mode enumeration."""
    IDLE = 'idle'
    VISIBILITY_ONLY = 'visibility_only'
    FULL = 'full'
    SELECTIVE = 'selective'


class VisibilityLevel(str, Enum, metaclass=IllumioEnumMeta):
    """Workload visibility level enumeration."""
    FLOW_FULL_DETAIL = 'flow_full_detail'
    FLOW_SUMMARY = 'flow_summary'
    FLOW_DROPS = 'flow_drops'
    FLOW_OFF = 'flow_off'
    ENHANCED_DATA_COLLECTION = 'enhanced_data_collection'


class PolicyDecision(str, Enum, metaclass=IllumioEnumMeta):
    """Traffic flow policy decision enumeration."""
    ALLOWED = 'allowed'
    BLOCKED = 'blocked'
    POTENTIALLY_BLOCKED = 'potentially_blocked'
    UNKNOWN = 'unknown'


class Transmission(str, Enum, metaclass=IllumioEnumMeta):
    """Traffic flow transmission enumeration."""
    BROADCAST = 'broadcast'
    MULTICAST = 'multicast'
    UNICAST = 'unicast'


class FlowDirection(str, Enum, metaclass=IllumioEnumMeta):
    """Traffic flow direction enumeration."""
    INBOUND = 'inbound'
    OUTBOUND = 'outbound'


class TrafficState(str, Enum, metaclass=IllumioEnumMeta):
    """Traffic flow state enumeration."""
    ACTIVE = 'active'
    CLOSED = 'closed'
    TIMED_OUT = 'timed out'
    SNAPSHOT = 'snapshot'
    NEW = 'new'
    UNKNOWN = 'unknown'
    INCOMPLETE = 'incomplete'


class ApplyTo(str, Enum, metaclass=IllumioEnumMeta):
    """Virtual service apply to value enumeration."""
    HOST_ONLY = 'host_only'
    INTERNAL_BRIDGE_NETWORK = 'internal_bridge_network'


class VENType(str, Enum, metaclass=IllumioEnumMeta):
    """VEN type enumeration."""
    SERVER = 'server'
    ENDPOINT = 'endpoint'
    CONTAINERIZED = 'containerized'


class ChangeType(str, Enum, metaclass=IllumioEnumMeta):
    """Resource event change type enumeration."""
    CREATE = 'create'
    UPDATE = 'update'
    DELETE = 'delete'


class EventSeverity(str, Enum, metaclass=IllumioEnumMeta):
    """Event severity enumeration."""
    EMERGENCY = 'emerg'
    ALERT = 'alert'
    CRITICAL = 'crit'
    ERROR = 'err'
    WARNING = 'warning'
    NOTICE = 'notice'
    INFO = 'info'
    DEBUG = 'debug'


class EventStatus(str, Enum, metaclass=IllumioEnumMeta):
    """Event status enumeration."""
    SUCCESS = 'success'
    FAILURE = 'failure'


__all__ = [
    'ACTIVE',
    'DRAFT',
    'AMS',
    'RESOLVE_AS_WORKLOADS',
    'RESOLVE_AS_VIRTUAL_SERVICES',
    'ANY_IP_LIST_NAME',
    'ALL_SERVICES_NAME',
    'PORT_MAX',
    'ICMP_CODE_MAX',
    'ICMP_TYPE_MAX',
    'FQDN_REGEX',
    'HREF_REGEX',
    'BULK_CHANGE_LIMIT',
    'PCE_APIS',
    'EnforcementMode',
    'LinkState',
    'EnforcementMode',
    'VisibilityLevel',
    'PolicyDecision',
    'Transmission',
    'FlowDirection',
    'TrafficState',
    'ApplyTo',
    'VENType',
    'ChangeType',
    'EventSeverity',
    'EventStatus',
]



# -*- coding: utf-8 -*-

"""This module provides classes related to security policy objects.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from .service import *
from .iplist import *
from .label import *
from .securityprincipal import *
from .virtualserver import *
from .virtualservice import *



def test_get_by_reference(pce, container_cluster):
    cluster = pce.container_clusters.get_by_reference(container_cluster.href)
    del container_cluster.container_cluster_token  # pop the token for the comparison
    assert cluster.href == container_cluster.href


def test_get_by_partial_name(pce, session_identifier, container_cluster):
    container_clusters = pce.container_clusters.get(params={'name': session_identifier})
    assert len(container_clusters) == 1


def test_update_container_cluster(pce, container_cluster):
    updated_description = 'Updated description'
    pce.container_clusters.update(container_cluster.href, {
        'description': updated_description
    })
    cluster = pce.container_clusters.get_by_reference(container_cluster.href)
    assert container_cluster.href == cluster.href
    assert cluster.description == updated_description
